# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/98_make_docs.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['MAGIC_COMMENTS', 'load_nb', 'repository_path', 'docs_path', 'save_nb', 'strip_metadata', 'docs_copy_nb',
           'list_notebooks', 'list_zips', 'docs_copy_dir', 'docs_copy_simulations', 'get_toc_part', 'make_docs']

# Internal Cell
#nbdev_comment from __future__ import annotations

import glob
import json
import os
import re
import secrets
import shutil
import subprocess
import sys
from typing import Dict, Union

from fastcore.imports import IN_IPYTHON

# Cell
MAGIC_COMMENTS = {
    "default_exp": "remove-cell",
    "exporti": "remove-cell",
    "export": "hide-input",
    "exports": None,
    "hide": "remove-cell",
    "hide_input": "remove-input",
    "hide_output": "remove-output",
    "collapse_input": "hide-input",
    "collapse_output": "hide-output",
}

# Cell
if IN_IPYTHON:
    ROOT = os.path.abspath('..')
else:
    ROOT = os.path.abspath('.')

# Cell
def load_nb(path: str) -> Dict:
    """ load a jupyter notebook as dictionary

    Args:
        path: the path of the notebook to load

    Returns:
        the notebook represented as a dictionary
    """
    with open(path, "r") as file:
        nb = json.load(file)
    return nb

# Cell
def repository_path(*path_parts: str, not_exist_ok: bool=False) -> str:
    """ Get and validate a path in the modelbuild repository

    Args:
        *path_parts: the path parts that will be joined together
            relative to the root of the repository.
        not_exist_ok: skip validation if True

    Returns:
        the absolute path of the referenced file.
    """
    if not (os.path.exists(path:=os.path.join(ROOT, "docs"))):
        raise FileNotFoundError(f"docs path {path!r} not found!")
    if not (os.path.exists(path:=os.path.join(ROOT, "nbs"))):
        raise FileNotFoundError(f"nbs path {path!r} not found!")
    if not (os.path.exists(path:=os.path.join(ROOT, "sax"))):
        raise FileNotFoundError(f"sax path {path!r} not found!")

    path = os.path.join(ROOT, *path_parts)
    if not not_exist_ok and not os.path.exists(path):
        raise FileNotFoundError(f"Path {path!r} does not exist.")

    return path

# Cell
def docs_path(*path_parts: str, not_exist_ok: bool=False) -> str:
    return repository_path('docs', *path_parts, not_exist_ok=not_exist_ok)

# Cell
def save_nb(nb: Dict, path: str) -> str:
    """ save a dictionary as a jupyter notebook

    Args:
        nb: the dictionary to convert into an ipynb file
        path: the path to save the notebook under

    Returns:
        the path where the notebook was saved.
    """
    path = os.path.abspath(path)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as file:
        json.dump(nb, file, indent=2)
    return path

# Cell
def strip_metadata(nb: Union[Dict, str]) -> Union[Dict,str]:
    path = ''
    if isinstance(nb, str):
        path = nb
        nb = load_nb(nb)
    for cell in nb['cells']:
        if not 'metadata' in cell:
            continue
        cell['metadata'] = {}
    if path:
        return save_nb(nb, path)
    return path

# Internal Cell
def iter_code_cells(nb):
    for cell in nb.get('cells', []):
        if not cell.get("cell_type", "") == "code":
            continue
        yield cell

# Internal Cell
def first_code_cell(nb):
    try:
        return next(iter_code_cells(nb))
    except StopIteration:
        return None

# Internal Cell
def get_default_exp(nb):
    first_cell = first_code_cell(nb) or {}
    first_source = first_cell.get('source', [])
    first_line = "" if not first_source else first_source[0]
    default_exp = first_line.split("default_exp")[-1].strip()
    return default_exp

# Internal Cell
def iter_function_names(source):
    for line in source:
        if not line.startswith("def "):
            continue
        if line.startswith("def _"):
            continue
        yield line.split("def ")[1].split("(")[0]

# Cell
def docs_copy_nb(relpath, docsrelpath=None):
    """copy a single notebook from src to dst with modified docs metadata."""
    src = repository_path(relpath)
    dst = docs_path((docsrelpath or relpath), not_exist_ok=True)
    nb = load_nb(src)
    nb_new = {**nb}
    nb_new["cells"] = []
    module = get_default_exp(nb)

    for cell in nb.get('cells', []):
        if not cell.get("cell_type", "") == "code":
            nb_new["cells"].append(cell)
            continue

        cell_tags = cell.get("metadata", {}).get("tags", [])
        source = cell.get("source") or [""]
        line = source[0].strip()

        if not line.startswith("#"):
            nb_new["cells"].append(cell)
            continue

        keys = [k.strip() for k in line.split(" ")]
        keys = [k for k in keys if k in MAGIC_COMMENTS]
        if keys:
            del source[0]
        for key in keys:
            tag = MAGIC_COMMENTS[key]
            if tag:
                cell_tags.append(tag)
        if len(cell_tags) > 0:
            cell["metadata"]["tags"] = cell_tags

        if not 'remove-cell' in cell_tags:
            for function_name in iter_function_names(source):
                extra_cell = {
                    "cell_type": "markdown",
                    "id": secrets.token_hex(8),
                    "metadata": {},
                    "source": [
                        ":::{eval-rst}\n",
                        f".. autofunction:: sax.{module}.{function_name}\n",
                        ":::\n"
                    ],
                }
                nb_new["cells"].append(extra_cell)

        cell["id"] = secrets.token_hex(8)
        nb_new["cells"].append(cell)

    os.makedirs(os.path.dirname(dst), exist_ok=True)
    save_nb(nb_new, dst)
    return dst

# Cell
def list_notebooks(dir):
    return glob.glob(os.path.join(dir, "**/*.ipynb"), recursive=True)

def list_zips(dir):
    return glob.glob(os.path.join(dir, "**/*.zip"), recursive=True)

# Cell
def docs_copy_dir(relpath):
    main_src = repository_path(relpath)
    for src in list_notebooks(main_src):
        rel = os.path.relpath(src, repository_path())
        docs_copy_nb(rel)
    for src in list_zips(main_src):
        rel = os.path.relpath(src, repository_path())
        dst = docs_path("_build", "html", rel, not_exist_ok=True)
        os.makedirs(os.path.dirname(dst), exist_ok=True)
        shutil.copy2(src, dst)

# Cell
def docs_copy_simulations():
    with_results, without_results = {}, {}
    for fn in os.listdir(simulations_path()):
        sim, ext = os.path.splitext(fn)
        if ext != ".ipynb":
            continue
        try:
            with_results[sim] = get_nominal_result(sim)
        except FileNotFoundError:
            without_results[sim] = {'hash': None, 'params': get_default_params(simulations_path(f"{sim}.ipynb"))}
        except ValueError:
            without_results[sim] = {'hash': None, 'params': get_default_params(simulations_path(f"{sim}.ipynb"))}
            #raise

    for k, r in with_results.items():
        docs_copy_nb(f"results/{k}/{r['hash']}/_simulated.ipynb", f"simulations/{k}.ipynb")

    for k, r in without_results.items():
        docs_copy_nb(f"simulations/{k}.ipynb", f"simulations/{k}.ipynb")

    shutil.copytree(simulations_path("img"), docs_path("_build", "html", "simulations", "img", not_exist_ok=True), dirs_exist_ok=True)
    return with_results, without_results

# Cell
def get_toc_part(toc, caption):
    parts = [p for p in toc["parts"] if caption == p["caption"]]
    try:
        return parts[0]
    except IndexError:
        raise ValueError(f"No TOC part with caption {caption!r} found.")

# Cell
def make_docs():
    docs_copy_nb("index.ipynb")
    docs_copy_dir("nbs")
    docs_copy_dir("examples")
    os.chdir(docs_path())
    subprocess.check_call([sys.executable.replace("python", "jupyter-book"), "build", "."])

# Cell
if __name__ == '__main__' and not IN_IPYTHON:
    make_docs()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/09a_nn_loss.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['mse', 'huber_loss', 'l2_reg']

# Cell
#nbdev_comment from __future__ import annotations

from typing import Dict

import jax.numpy as jnp
from ..typing_ import ComplexFloat

# Cell

def mse(x: ComplexFloat, y: ComplexFloat) -> float:
    """mean squared error"""
    return ((x - y) ** 2).mean()

# Cell

def huber_loss(x: ComplexFloat, y: ComplexFloat, delta: float=0.5) -> float:
    """huber loss"""
    return ((delta ** 2) * ((1.0 + ((x - y) / delta) ** 2) ** 0.5 - 1.0)).mean()

# Cell

def l2_reg(weights: Dict[str, ComplexFloat]) -> float:
    """L2 regularization loss"""
    numel = 0
    loss = 0.0
    for w in (v for k, v in weights.items() if k[0] in ("w", "b")):
        numel = numel + w.size
        loss = loss + (jnp.abs(w) ** 2).sum()
    return loss / numel


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_utils.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['block_diag', 'clean_string', 'copy_settings', 'validate_settings', 'try_float', 'flatten_dict',
           'unflatten_dict', 'get_ports', 'get_port_combinations', 'get_settings', 'grouped_interp', 'merge_dicts',
           'mode_combinations', 'reciprocal', 'rename_params', 'rename_ports', 'update_settings',
           'validate_not_mixedmode', 'validate_multimode', 'validate_sdict', 'get_inputs_outputs', 'hash_dict']

# Cell
#nbdev_comment from __future__ import annotations

import inspect
import re
import warnings
from functools import lru_cache, partial, wraps
from hashlib import md5
from typing import Any, Callable, Dict, Iterable, Iterator, Tuple, Union, cast, overload

import numpy as np
import orjson
from natsort import natsorted
from .typing_ import (
    Array,
    ComplexFloat,
    Float,
    Model,
    ModelFactory,
    SCoo,
    SDense,
    SDict,
    Settings,
    SType,
    is_mixedmode,
    is_model,
    is_model_factory,
    is_scoo,
    is_sdense,
    is_sdict,
)

try:
    import jax
    import jax.numpy as jnp
    import jax.scipy as jsp
    import jax.scipy.linalg
    JAX_AVAILABLE = True
except ImportError:
    import numpy as jnp
    import scipy as jsp
    import scipy.linalg
    JAX_AVAILABLE = False

# Cell
def block_diag(*arrs: Array) -> Array:
    """create block diagonal matrix with arbitrary batch dimensions """
    batch_shape = arrs[0].shape[:-2]

    B = jnp.prod(jnp.asarray(batch_shape, dtype=int))

    N = 0
    for arr in arrs:
        if batch_shape != arr.shape[:-2]:
            raise ValueError("batch dimensions for given arrays don't match.")
        m, n = arr.shape[-2:]
        if m != n:
            raise ValueError("given arrays are not square.")
        N += n

    arrs = [arr.reshape(-1, arr.shape[-2], arr.shape[-1]) for arr in arrs]

    if JAX_AVAILABLE:
        batch_block_diag = jax.vmap(jsp.linalg.block_diag, in_axes=0, out_axes=0)
        block_diag = batch_block_diag(*arrs)
    else:
        arrs = [jsp.linalg.block_diag(*[arr[i] for arr in arrs]) for i in range(B)]
        block_diag = jnp.stack(arrs, 0)

    return block_diag.reshape(*batch_shape, N, N)

# Cell

def clean_string(s: str, dot='p', minus='m', other='_') -> str:
    """clean a string such that it is a valid python identifier"""
    original = s
    s = s.strip()
    s = s.replace(".", dot)  # dot
    s = s.replace("-", minus)  # minus
    s = re.sub("[^0-9a-zA-Z]", other, s)
    if s[0] in "0123456789":
        s = "_" + s
    if s != original:
        warnings.warn(f"modified string {original} in an attempt to make valid python identifier: {s}")
    if not s.isidentifier():
        raise ValueError(f"failed to clean string to a valid python identifier: {s}")
    return s

# Cell

def copy_settings(settings: Settings) -> Settings:
    """copy a parameter dictionary"""
    return validate_settings(settings)  # validation also copies

def validate_settings(settings: Settings) -> Settings:
    """Validate a parameter dictionary"""
    _settings = {}
    for k, v in settings.items():
        if isinstance(v, dict):
            _settings[k] = validate_settings(v)
        else:
            _settings[k] = try_float(v)
    return _settings

def try_float(f: Any) -> Any:
    """try converting an object to float, return unchanged object on fail"""
    try:
        return jnp.asarray(f, dtype=float)
    except (ValueError, TypeError):
        return f

# Cell
def flatten_dict(dic: Dict[str, Any], sep: str = ",") -> Dict[str, Any]:
    """flatten a nested dictionary"""
    return _flatten_dict(dic, sep=sep)


def _flatten_dict(
    dic: Dict[str, Any], sep: str = ",", frozen: bool = False, parent_key: str = ""
) -> Dict[str, Any]:
    items = []
    for k, v in dic.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(
                _flatten_dict(v, sep=sep, frozen=frozen, parent_key=new_key).items()
            )
        else:
            items.append((new_key, v))

    return dict(items)

# Cell

def unflatten_dict(dic, sep=","):
    """unflatten a flattened dictionary """

    # from: https://gist.github.com/fmder/494aaa2dd6f8c428cede
    items = dict()

    for k, v in dic.items():
        keys = k.split(sep)
        sub_items = items
        for ki in keys[:-1]:
            if ki in sub_items:
                sub_items = sub_items[ki]
            else:
                sub_items[ki] = dict()
                sub_items = sub_items[ki]

        sub_items[keys[-1]] = v

    return items

# Cell

def get_ports(S: Union[Model, SType]) -> Tuple[str, ...]:
    """get port names of a model or an stype"""
    if is_model(S):
        return _get_ports_from_model(cast(Model, S))
    elif is_sdict(S):
        ports_set = {p1 for p1, _ in S} | {p2 for _, p2 in S}
        return tuple(natsorted(ports_set))
    elif is_scoo(S) or is_sdense(S):
        *_, ports_map = S
        return tuple(natsorted(ports_map.keys()))
    else:
        raise ValueError("Could not extract ports for given S")

@lru_cache(maxsize=4096)  # cache to prevent future tracing
def _get_ports_from_model(model: Model) -> Tuple[str, ...]:
    if JAX_AVAILABLE:
        S: SType = jax.eval_shape(model)
    else:
        warnings.warn("[NO JAX] requesting the ports of a model requires evaluating the model!")
        S = model()
    return get_ports(S)

# Cell

def get_port_combinations(S: Union[Model, SType]) -> Tuple[Tuple[str, str], ...]:
    """get port combinations of a model or an stype"""

    if is_model(S):
        S = cast(Model, S)
        return _get_port_combinations_from_model(S)
    elif is_sdict(S):
        S = cast(SDict, S)
        return tuple(S.keys())
    elif is_scoo(S):
        Si, Sj, _, pm = cast(SCoo, S)
        rpm = {int(i): str(p) for p, i in pm.items()}
        return tuple(natsorted((rpm[int(i)], rpm[int(j)]) for i, j in zip(Si, Sj)))
    elif is_sdense(S):
        _, pm = cast(SDense, S)
        return tuple(natsorted((p1, p2) for p1 in pm for p2 in pm))
    else:
        raise ValueError("Could not extract ports for given S")

@lru_cache(maxsize=4096)  # cache to prevent future tracing
def _get_port_combinations_from_model(model: Model) -> Tuple[Tuple[str, str], ...]:
    if JAX_AVAILABLE:
        S: SType = jax.eval_shape(model)
    else:
        warnings.warn("[NO JAX] requesting the ports of a model requires evaluating the model!")
        S = model()
    return get_port_combinations(S)

# Cell
def get_settings(model: Union[Model, ModelFactory]) -> Settings:
    """Get the parameters of a SAX model function"""

    signature = inspect.signature(model)

    settings: Settings = {
        k: (v.default if not isinstance(v, dict) else v)
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

    # make sure an inplace operation of resulting dict does not change the
    # circuit parameters themselves
    return copy_settings(settings)

# Cell
def grouped_interp(wl: Float, wls: Float, phis: Float) -> Float:
    """Grouped phase interpolation"""
    if not JAX_AVAILABLE:
        raise NotImplementedError("[NO JAX] grouped_interp function not implemented when JAX not available. Please install JAX.")

    @partial(jax.vmap, in_axes=(0, None, None), out_axes=0)
    @jax.jit
    def _grouped_interp(
        wl: Array,  # 0D array (not-vmapped) ; 1D array (vmapped)
        wls: Array,  # 1D array
        phis: Array,  # 1D array
    ) -> Array:
        dphi_dwl = (phis[1::2] - phis[::2]) / (wls[1::2] - wls[::2])
        phis = phis[::2]
        wls = wls[::2]
        dwl = (wls[1:] - wls[:-1]).mean(0, keepdims=True)

        t = (wl - wls + 1e-5 * dwl) / dwl  # small offset to ensure no values are zero
        t = jnp.where(jnp.abs(t) < 1, t, 0)
        m0 = jnp.where(t > 0, size=1)[0]
        m1 = jnp.where(t < 0, size=1)[0]
        t = t[m0]
        wl0 = wls[m0]
        wl1 = wls[m1]
        phi0 = phis[m0]
        phi1 = phis[m1]
        dphi_dwl0 = dphi_dwl[m0]
        dphi_dwl1 = dphi_dwl[m1]
        _phi0 = phi0 - 0.5 * (wl1 - wl0) * (
            dphi_dwl0 * (t ** 2 - 2 * t) - dphi_dwl1 * t ** 2
        )
        _phi1 = phi1 - 0.5 * (wl1 - wl0) * (
            dphi_dwl0 * (t - 1) ** 2 - dphi_dwl1 * (t ** 2 - 1)
        )
        phis = jnp.arctan2(
            (1 - t) * jnp.sin(_phi0) + t * jnp.sin(_phi1),
            (1 - t) * jnp.cos(_phi0) + t * jnp.cos(_phi1),
        )
        return phis

    wl = cast(Array, jnp.asarray(wl))
    wls = cast(Array, jnp.asarray(wls))
    # make sure values between -pi and pi
    phis = cast(Array, jnp.asarray(phis)) % (2 * jnp.pi)
    phis = jnp.where(phis > jnp.pi, phis - 2 * jnp.pi, phis)
    if not wls.ndim == 1:
        raise ValueError("grouped_interp: wls should be a 1D array")
    if not phis.ndim == 1:
        raise ValueError("grouped_interp: wls should be a 1D array")
    if not wls.shape == phis.shape:
        raise ValueError("grouped_interp: wls and phis shape does not match")
    return _grouped_interp(wl.reshape(-1), wls, phis).reshape(*wl.shape)

# Cell
def merge_dicts(*dicts: Dict) -> Dict:
    """merge (possibly deeply nested) dictionaries"""
    if len(dicts) == 1:
        return dict(_generate_merged_dict(dicts[0], {}))
    elif len(dicts) == 2:
        return dict(_generate_merged_dict(dicts[0], dicts[1]))
    else:
        return merge_dicts(dicts[0], merge_dicts(*dicts[1:]))


def _generate_merged_dict(dict1: Dict, dict2: Dict) -> Iterator[Tuple[Any, Any]]:
    # inspired by https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries
    keys = {**{k: None for k in dict1}, **{k: None for k in dict2}} # keep key order, values irrelevant
    for k in keys:
        if k in dict1 and k in dict2:
            v1, v2 = dict1[k], dict2[k]
            if isinstance(v1, dict) and isinstance(v2, dict):
                v = dict(_generate_merged_dict(v1, v2))
            else:
                # If one of the values is not a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                v = v2
        elif k in dict1:
            v = dict1[k]
        else: # k in dict2:
            v = dict2[k]

        if isinstance(v, dict):
            yield (k, {**v}) # shallow copy of dict
        else:
            yield (k, v)

# Cell

def mode_combinations(
    modes: Iterable[str], cross: bool = False
) -> Tuple[Tuple[str, str], ...]:
    """create mode combinations for a collection of given modes"""
    if cross:
        mode_combinations = natsorted((m1, m2) for m1 in modes for m2 in modes)
    else:
        mode_combinations = natsorted((m, m) for m in modes)
    return tuple(mode_combinations)

# Cell
def reciprocal(sdict: SDict) -> SDict:
    """Make an SDict reciprocal"""
    if is_sdict(sdict):
        return {
            **{(p1, p2): v for (p1, p2), v in sdict.items()},
            **{(p2, p1): v for (p1, p2), v in sdict.items()},
        }
    else:
        raise ValueError("sax.reciprocal is only valid for SDict types")

# Internal Cell

@overload
def rename_params(model: ModelFactory, renamings: Dict[str, str]) -> ModelFactory:
    ...


@overload
def rename_params(model: Model, renamings: Dict[str, str]) -> Model:
    ...

# Cell
def rename_params(
    model: Union[Model, ModelFactory], renamings: Dict[str, str]
) -> Union[Model, ModelFactory]:
    """rename the parameters of a `Model` or `ModelFactory` given a renamings mapping old parameter names to new."""

    reversed_renamings = {v: k for k, v in renamings.items()}
    if len(reversed_renamings) < len(renamings):
        raise ValueError("Multiple old names point to the same new name!")

    if is_model_factory(model):
        old_model_factory = cast(ModelFactory, model)
        old_settings = get_settings(model)

        @wraps(old_model_factory)
        def new_model_factory(**settings):
            old_settings = {
                reversed_renamings.get(k, k): v for k, v in settings.items()
            }
            model = old_model_factory(**old_settings)
            return rename_params(model, renamings)

        new_settings = {renamings.get(k, k): v for k, v in old_settings.items()}
        _replace_kwargs(new_model_factory, **new_settings)

        return new_model_factory

    elif is_model(model):
        old_model = cast(Model, model)
        old_settings = get_settings(model)

        @wraps(old_model)
        def new_model(**settings):
            old_settings = {
                reversed_renamings.get(k, k): v for k, v in settings.items()
            }
            return old_model(**old_settings)

        new_settings = {renamings.get(k, k): v for k, v in old_settings.items()}
        _replace_kwargs(new_model, **new_settings)

        return new_model

    else:
        raise ValueError(
            "rename_params should be used to decorate a Model or ModelFactory."
        )

def _replace_kwargs(func: Callable, **kwargs: ComplexFloat):
    """Change the kwargs signature of a function"""
    sig = inspect.signature(func)
    settings = [
        inspect.Parameter(k, inspect.Parameter.KEYWORD_ONLY, default=v)
        for k, v in kwargs.items()
    ]
    func.__signature__ = sig.replace(parameters=settings)

# Internal Cell

@overload
def rename_ports(S: SDict, renamings: Dict[str, str]) -> SDict:
    ...


@overload
def rename_ports(S: SCoo, renamings: Dict[str, str]) -> SCoo:
    ...


@overload
def rename_ports(S: SDense, renamings: Dict[str, str]) -> SDense:
    ...


@overload
def rename_ports(S: Model, renamings: Dict[str, str]) -> Model:
    ...


@overload
def rename_ports(S: ModelFactory, renamings: Dict[str, str]) -> ModelFactory:
    ...

# Cell

def rename_ports(
    S: Union[SType, Model, ModelFactory], renamings: Dict[str, str]
) -> Union[SType, Model, ModelFactory]:
    """rename the ports of an `SDict`, `Model` or `ModelFactory` given a renamings mapping old port names to new."""
    if is_scoo(S):
        Si, Sj, Sx, ports_map = cast(SCoo, S)
        ports_map = {renamings[p]: i for p, i in ports_map.items()}
        return Si, Sj, Sx, ports_map
    elif is_sdense(S):
        Sx, ports_map = cast(SDense, S)
        ports_map = {renamings[p]: i for p, i in ports_map.items()}
        return Sx, ports_map
    elif is_sdict(S):
        sdict = cast(SDict, S)
        original_ports = get_ports(sdict)
        assert len(renamings) == len(original_ports)
        return {(renamings[p1], renamings[p2]): v for (p1, p2), v in sdict.items()}
    elif is_model(S):
        old_model = cast(Model, S)

        @wraps(old_model)
        def new_model(**settings) -> SType:
            return rename_ports(old_model(**settings), renamings)

        return new_model
    elif is_model_factory(S):
        old_model_factory = cast(ModelFactory, S)

        @wraps(old_model_factory)
        def new_model_factory(**settings) -> Callable[..., SType]:
            return rename_ports(old_model_factory(**settings), renamings)

        return new_model_factory
    else:
        raise ValueError("Cannot rename ports for type {type(S)}")

# Cell

def update_settings(
    settings: Settings, *compnames: str, **kwargs: ComplexFloat
) -> Settings:
    """update a nested settings dictionary"""
    _settings = {}
    if not compnames:
        for k, v in settings.items():
            if isinstance(v, dict):
                _settings[k] = update_settings(v, **kwargs)
            else:
                if k in kwargs:
                    _settings[k] = try_float(kwargs[k])
                else:
                    _settings[k] = try_float(v)
    else:
        for k, v in settings.items():
            if isinstance(v, dict):
                if k == compnames[0]:
                    _settings[k] = update_settings(v, *compnames[1:], **kwargs)
                else:
                    _settings[k] = v
            else:
                _settings[k] = try_float(v)
    return _settings

# Cell
def validate_not_mixedmode(S: SType):
    """validate that an stype is not 'mixed mode' (i.e. invalid)

    Args:
        S: the stype to validate
    """

    if is_mixedmode(S):  # mixed mode
        raise ValueError(
            "Given SType is neither multimode or singlemode. Please check the port "
            "names: they should either ALL contain the '@' separator (multimode) "
            "or NONE should contain the '@' separator (singlemode)."
        )

# Cell
def validate_multimode(S: SType, modes=("te", "tm")) -> None:
    """validate that an stype is multimode and that the given modes are present."""
    try:
        current_modes = set(p.split("@")[1] for p in get_ports(S))
    except IndexError:
        raise ValueError("The given stype is not multimode.")
    for mode in modes:
        if mode not in current_modes:
            raise ValueError(
                f"Could not find mode '{mode}' in one of the multimode models."
            )

# Cell
def validate_sdict(sdict: Any) -> None:
    """Validate an `SDict`"""

    if not isinstance(sdict, dict):
        raise ValueError("An SDict should be a dictionary.")
    for ports in sdict:
        if not isinstance(ports, tuple) and not len(ports) == 2:
            raise ValueError(f"SDict keys should be length-2 tuples. Got {ports}")
        p1, p2 = ports
        if not isinstance(p1, str) or not isinstance(p2, str):
            raise ValueError(
                f"SDict ports should be strings. Got {ports} "
                f"({type(ports[0])}, {type(ports[1])})"
            )

# Cell

def get_inputs_outputs(ports: Tuple[str, ...]):
    inputs = tuple(p for p in ports if p.lower().startswith("in"))
    outputs = tuple(p for p in ports if not p.lower().startswith("in"))
    if not inputs:
        inputs = tuple(p for p in ports if not p.lower().startswith("out"))
        outputs = tuple(p for p in ports if p.lower().startswith("out"))
    return inputs, outputs

# Cell

def hash_dict(dic: Dict) -> int:
    return int(
        md5(
            orjson.dumps(_numpyfy(dic), option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SORT_KEYS)
        ).hexdigest(),
        16,
    )

def _numpyfy(obj: Any):
    if not isinstance(obj, dict):
        return np.asarray(obj)
    else:
        return {k: _numpyfy(v) for k, v in obj.items()}


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/09c_nn_core.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['preprocess', 'dense', 'generate_dense_weights']

# Cell
#nbdev_comment from __future__ import annotations

from typing import Callable, Dict, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from .utils import denormalize, normalize
from ..typing_ import Array, ComplexFloat

# Cell
def preprocess(*params: ComplexFloat) -> ComplexFloat:
    """preprocess parameters

    > Note: (1) all arguments are first casted into the same shape. (2) then pairs
      of arguments are divided into each other to create relative arguments. (3) all
      arguments are then stacked into one big tensor
    """
    x = jnp.stack(jnp.broadcast_arrays(*params), -1)
    assert isinstance(x, jnp.ndarray)
    to_concatenate = [x]
    for i in range(1, x.shape[-1]):
        _x = jnp.roll(x, shift=i, axis=-1)
        to_concatenate.append(x / _x)
        to_concatenate.append(_x / x)
    x = jnp.concatenate(to_concatenate, -1)
    assert isinstance(x, jnp.ndarray)
    return x

# Cell
def dense(
    weights: Dict[str, Array],
    *params: ComplexFloat,
    x_norm: Tuple[float, float] = (0.0, 1.0),
    y_norm: Tuple[float, float] = (0.0, 1.0),
    preprocess: Callable = preprocess,
    activation: Callable = jax.nn.leaky_relu,
) -> ComplexFloat:
    """simple dense neural network"""
    x_mean, x_std = x_norm
    y_mean, y_std = y_norm
    x = preprocess(*params)
    x = normalize(x, mean=x_mean, std=x_std)
    for i in range(len([w for w in weights if w.startswith("w")])):
        x = activation(x @ weights[f"w{i}"] + weights.get(f"b{i}", 0.0))
    y = denormalize(x, mean=y_mean, std=y_std)
    return y

# Cell
def generate_dense_weights(
    key: Union[int, Array],
    sizes: Tuple[int, ...],
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    preprocess=preprocess,
) -> Dict[str, ComplexFloat]:
    """Generate the weights for a dense neural network"""

    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
    assert isinstance(key, jnp.ndarray)

    sizes = tuple(s for s in sizes)
    if input_names:
        arr = preprocess(*jnp.ones(len(input_names)))
        assert isinstance(arr, jnp.ndarray)
        sizes = (arr.shape[-1],) + sizes
    if output_names:
        sizes = sizes + (len(output_names),)

    keys = jax.random.split(key, 2 * len(sizes))
    rand = jax.nn.initializers.lecun_normal()
    weights = {}
    for i, (m, n) in enumerate(zip(sizes[:-1], sizes[1:])):
        weights[f"w{i}"] = rand(keys[2 * i], (m, n))
        weights[f"b{i}"] = rand(keys[2 * i + 1], (1, n)).ravel()

    return weights


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_patched.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = []

# Cell
#nbdev_comment from __future__ import annotations

import re
from textwrap import dedent

from fastcore.basics import patch_to
from .typing_ import is_complex_float, is_float

try:
    import jax.numpy as jnp
    from flax.core import FrozenDict
    from jaxlib.xla_extension import DeviceArray
    JAX_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    import numpy as jnp
    DeviceArray = jnp.ndarray
    FrozenDict = dict
    JAX_AVAILABLE = False

# Internal Cell

if JAX_AVAILABLE:
    @patch_to(FrozenDict)
    def __repr__(self):  # type: ignore
        _dict = lambda d: dict(
            {k: (v if not isinstance(v, self.__class__) else dict(v)) for k, v in d.items()}
        )
        return f"{self.__class__.__name__}({dict.__repr__(_dict(self))})"

# Internal Cell
if JAX_AVAILABLE:
    @patch_to(DeviceArray)
    def __repr__(self):  # type: ignore
        if self.ndim == 0 and is_float(self):
            v = float(self)
            return repr(round(v, 5)) if abs(v) > 1e-4 else repr(v)
        elif self.ndim == 0 and is_complex_float(self):
            r, i = float(self.real), float(self.imag)
            r = round(r, 5) if abs(r) > 1e-4 else r
            i = round(i, 5) if abs(i) > 1e-4 else i
            s = repr(r + 1j * i)
            if s[0] == "(" and s[-1] == ")":
                s = s[1:-1]
            return s
        else:
            s = super(self.__class__, self).__repr__()
            s = s.replace("DeviceArray(", "      array(")
            s = re.sub(r", dtype=.*[,)]", "", s)
            s = re.sub(r" weak_type=.*[,)]", "", s)
            return dedent(s)+")"


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06a_netlist.ipynb (unless otherwise specified).

__all__ = ['Component', 'PortEnum', 'Placement', 'Route', 'Netlist', 'RecursiveNetlist', 'netlist', 'load_netlist',
           'load_recursive_netlist', 'get_netlist_instances_by_prefix', 'get_component_instances',
           'get_netlist_instances_by_prefix', 'get_component_instances']

# Cell
import os
import re
from enum import Enum
from functools import lru_cache, partial
from typing import Any, Callable, Dict, Optional, Union

import black
import numpy as np
import yaml
from pydantic import BaseModel as _BaseModel, RootModel
from pydantic import ConfigDict, Field, ValidationError, field_validator
from .utils import clean_string, get_settings, hash_dict

# Internal Cell

class BaseModel(_BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
    )

    def __repr__(self):
        s = super().__repr__()
        s = black.format_str(s, mode=black.Mode())
        return s

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash_dict(self.dict())

# Cell

class Component(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
    )

    component: Union[str, Dict[str, Any]] = Field(..., title="Component")
    settings: Optional[Dict[str, Any]] = Field(None, title="Settings")

    # this was added:
    @field_validator("component")
    @classmethod
    def validate_component_name(cls, value):
        if "," in value:
            raise ValueError(
                f"Invalid component string. Should not contain ','. Got: {value}"
            )
        return clean_string(value)


class PortEnum(Enum):
    ce = "ce"
    cw = "cw"
    nc = "nc"
    ne = "ne"
    nw = "nw"
    sc = "sc"
    se = "se"
    sw = "sw"
    center = "center"
    cc = "cc"


class Placement(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
    )

    x: Optional[Union[str, float]] = Field(0, title="X")
    y: Optional[Union[str, float]] = Field(0, title="Y")
    xmin: Optional[Union[str, float]] = Field(None, title="Xmin")
    ymin: Optional[Union[str, float]] = Field(None, title="Ymin")
    xmax: Optional[Union[str, float]] = Field(None, title="Xmax")
    ymax: Optional[Union[str, float]] = Field(None, title="Ymax")
    dx: Optional[float] = Field(0, title="Dx")
    dy: Optional[float] = Field(0, title="Dy")
    port: Optional[Union[str, PortEnum]] = Field(None, title="Port")
    rotation: Optional[int] = Field(0, title="Rotation")
    mirror: Optional[bool] = Field(False, title="Mirror")


class Route(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
    )

    links: Dict[str, str] = Field(..., title="Links")
    settings: Optional[Dict[str, Any]] = Field(None, title="Settings")
    routing_strategy: Optional[str] = Field(None, title="Routing Strategy")


class Netlist(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
    )

    instances: Dict[str, Component] = Field(..., title="Instances")
    connections: Optional[Dict[str, str]] = Field(None, title="Connections")
    ports: Optional[Dict[str, str]] = Field(None, title="Ports")
    placements: Optional[Dict[str, Placement]] = Field(None, title="Placements")

    # these were removed (irrelevant for SAX):

    # routes: Optional[Dict[str, Route]] = Field(None, title='Routes')
    # name: Optional[str] = Field(None, title='Name')
    # info: Optional[Dict[str, Any]] = Field(None, title='Info')
    # settings: Optional[Dict[str, Any]] = Field(None, title='Settings')
    # pdk: Optional[str] = Field(None, title='Pdk')

    # these are extra additions:

    @field_validator("instances", mode="before")
    @classmethod
    def coerce_different_type_instance_into_component_model(cls, instances):
        new_instances = {}
        for k, v in instances.items():
            if isinstance(v, str):
                v = {
                    "component": v,
                    "settings": {},
                }
            new_instances[k] = v

        return new_instances

    @staticmethod
    def clean_instance_string(value):
        if "," in value:
            raise ValueError(
                f"Invalid instance string. Should not contain ','. Got: {value}"
            )
        return clean_string(value)

    @field_validator("instances")
    @classmethod
    def validate_instance_names(cls, instances):
        return {cls.clean_instance_string(k): v for k, v in instances.items()}

    @field_validator("placements")
    @classmethod
    def validate_placement_names(cls, placements):
        return {cls.clean_instance_string(k): v for k, v in placements.items()}

    @classmethod
    def clean_connection_string(cls, value):
        *comp, port = value.split(",")
        comp = cls.clean_instance_string(",".join(comp))
        return f"{comp},{port}"

    @field_validator("connections")
    @classmethod
    def validate_connection_names(cls, connections):
        return {
            cls.clean_connection_string(k): cls.clean_connection_string(v)
            for k, v in connections.items()
        }

    @field_validator("ports")
    @classmethod
    def validate_port_names(cls, ports):
        return {
            cls.clean_instance_string(k): cls.clean_connection_string(v)
            for k, v in ports.items()
        }

# Cell

class RecursiveNetlist(RootModel):
    model_config = ConfigDict(extra="ignore", frozen=True)
    root: Dict[str, Netlist]

# Cell

def netlist(dic: Dict) -> RecursiveNetlist:
    if isinstance(dic, RecursiveNetlist):
        return dic
    elif isinstance(dic, Netlist):
        dic = dic.dict()
    try:
        flat_net = Netlist.model_validate(dic)
        net = RecursiveNetlist.model_validate({'top_level': flat_net})
    except ValidationError:
        net = RecursiveNetlist.model_validate(dic)
    return net

# Cell
@lru_cache()
def load_netlist(pic_path) -> Netlist:
    with open(pic_path, "r") as file:
        net = yaml.safe_load(file.read())
    return Netlist.model_validate(net)

# Cell
@lru_cache()
def load_recursive_netlist(pic_path, ext='.yml'):
    folder_path = os.path.dirname(os.path.abspath(pic_path))
    _clean_string = lambda path: clean_string(re.sub(ext, "", os.path.split(path)[-1]))
    netlists = {_clean_string(pic_path): None} # the circuit we're interested in should come first.
    for filename in os.listdir(folder_path):
        path = os.path.join(folder_path, filename)
        if not os.path.isfile(path) or not path.endswith(ext):
            continue
        netlists[_clean_string(path)] = load_netlist(path)
    return RecursiveNetlist.model_validate(netlists)

# Cell
def get_netlist_instances_by_prefix(
        recursive_netlist: RecursiveNetlist,
        prefix: str,
):
    """
    Returns a list of all instances with a given prefix in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        prefix: The prefix to search for.

    Returns:
        A list of all instances with the given prefix.
    """
    recursive_netlist_root = recursive_netlist.dict()["__root__"]
    result = []
    for key in recursive_netlist_root.keys():
        if key.startswith(prefix):
            result.append(key)
    return result

# Cell
def get_component_instances(
        recursive_netlist: RecursiveNetlist,
        top_level_prefix: str,
        component_name_prefix: str,
):
    """
    Returns a dictionary of all instances of a given component in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        top_level_prefix: The prefix of the top level instance.
        component_name_prefix: The name of the component to search for.

    Returns:
        A dictionary of all instances of the given component.
    """
    instance_names = []
    recursive_netlist_root = recursive_netlist.dict()["__root__"]
    top_level_prefix = get_netlist_instances_by_prefix(recursive_netlist, prefix=top_level_prefix)[
        0
    ]  # Should only be one in a netlist-to-digraph. Can always be very specified.
    for key in recursive_netlist_root[top_level_prefix]["instances"]:
        if recursive_netlist_root[top_level_prefix]["instances"][key]["component"].startswith(component_name_prefix):
            # Note priority encoding on match.
            instance_names.append(key)
    return {component_name_prefix: instance_names}

# Cell
def get_netlist_instances_by_prefix(
        recursive_netlist: RecursiveNetlist,
        prefix: str,
):
    """
    Returns a list of all instances with a given prefix in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        prefix: The prefix to search for.

    Returns:
        A list of all instances with the given prefix.
    """
    recursive_netlist_root = recursive_netlist.dict()["__root__"]
    result = []
    for key in recursive_netlist_root.keys():
        if key.startswith(prefix):
            result.append(key)
    return result

# Cell
def get_component_instances(
        recursive_netlist: RecursiveNetlist,
        top_level_prefix: str,
        component_name_prefix: str,
):
    """
    Returns a dictionary of all instances of a given component in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        top_level_prefix: The prefix of the top level instance.
        component_name_prefix: The name of the component to search for.

    Returns:
        A dictionary of all instances of the given component.
    """
    instance_names = []
    recursive_netlist_root = recursive_netlist.dict()["__root__"]
    top_level_prefix = get_netlist_instances_by_prefix(recursive_netlist, prefix=top_level_prefix)[
        0
    ]  # Should only be one in a netlist-to-digraph. Can always be very specified.
    for key in recursive_netlist_root[top_level_prefix]["instances"]:
        if recursive_netlist_root[top_level_prefix]["instances"][key]["component"].startswith(component_name_prefix):
            # Note priority encoding on match.
            instance_names.append(key)
    return {component_name_prefix: instance_names}


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/09d_nn_io.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['load_nn_weights_json', 'save_nn_weights_json', 'get_available_sizes', 'get_dense_weights_path',
           'get_norm_path', 'load_nn_dense']

# Cell
#nbdev_comment from __future__ import annotations

import json
import os
import re
from typing import Callable, Dict, List, Optional, Tuple

import jax.numpy as jnp
from .core import dense, preprocess
from .utils import norm
from ..typing_ import ComplexFloat

# Cell

def load_nn_weights_json(path: str) -> Dict[str, ComplexFloat]:
    """Load json weights from given path"""
    path = os.path.abspath(os.path.expanduser(path))
    weights = {}
    if os.path.exists(path):
        with open(path, "r") as file:
            for k, v in json.load(file).items():
                _v = jnp.array(v, dtype=float)
                assert isinstance(_v, jnp.ndarray)
                weights[k] = _v
    return weights

# Cell
def save_nn_weights_json(weights: Dict[str, ComplexFloat], path: str):
    """Save json weights to given path"""
    path = os.path.abspath(os.path.expanduser(path))
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as file:
        _weights = {}
        for k, v in weights.items():
            v = jnp.atleast_1d(jnp.array(v))
            assert isinstance(v, jnp.ndarray)
            _weights[k] = v.tolist()
        json.dump(_weights, file)

# Cell
def get_available_sizes(
    dirpath: str,
    prefix: str,
    input_names: Tuple[str, ...],
    output_names: Tuple[str, ...],
) -> List[Tuple[int, ...]]:
    """Get all available json weight hidden sizes given filename parameters

    > Note: this function does NOT return the input size and the output size
      of the neural network. ONLY the hidden sizes are reported. The input
      and output sizes can easily be derived from `input_names` (after
      preprocessing) and `output_names`.
    """
    all_weightfiles = os.listdir(dirpath)
    possible_weightfiles = (
        s for s in all_weightfiles if s.endswith(f"-{'-'.join(output_names)}.json")
    )
    possible_weightfiles = (
        s
        for s in possible_weightfiles
        if s.startswith(f"{prefix}-{'-'.join(input_names)}")
    )
    possible_weightfiles = (re.sub("[^0-9x]", "", s) for s in possible_weightfiles)
    possible_weightfiles = (re.sub("^x*", "", s) for s in possible_weightfiles)
    possible_weightfiles = (re.sub("x[^0-9]*$", "", s) for s in possible_weightfiles)
    possible_hidden_sizes = (s.strip() for s in possible_weightfiles if s.strip())
    possible_hidden_sizes = (
        tuple(hs.strip() for hs in s.split("x") if hs.strip())
        for s in possible_hidden_sizes
    )
    possible_hidden_sizes = (
        tuple(int(hs) for hs in s[1:-1]) for s in possible_hidden_sizes if len(s) > 2
    )
    possible_hidden_sizes = sorted(
        possible_hidden_sizes, key=lambda hs: (len(hs), max(hs))
    )
    return possible_hidden_sizes

# Cell

def get_dense_weights_path(
    *sizes: int,
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    dirpath: str = "weights",
    prefix: str = "dense",
    preprocess=preprocess,
):
    """Create the SAX conventional path for a given weight dictionary"""
    if input_names:
        num_inputs = preprocess(*jnp.ones(len(input_names))).shape[0]
        sizes = (num_inputs,) + sizes
    if output_names:
        sizes = sizes + (len(output_names),)
    path = os.path.abspath(os.path.join(dirpath, prefix))
    if input_names:
        path = f"{path}-{'-'.join(input_names)}"
    if sizes:
        path = f"{path}-{'x'.join(str(s) for s in sizes)}"
    if output_names:
        path = f"{path}-{'-'.join(output_names)}"
    return f"{path}.json"

# Cell

def get_norm_path(
    *shape: int,
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    dirpath: str = "norms",
    prefix: str = "norm",
    preprocess=preprocess,
):
    """Create the SAX conventional path for the normalization constants"""
    if input_names and output_names:
        raise ValueError(
            "To get the norm name, one can only specify `input_names` OR `output_names`."
        )
    if input_names:
        num_inputs = preprocess(*jnp.ones(len(input_names))).shape[0]
        shape = (num_inputs,) + shape
    if output_names:
        shape = shape + (len(output_names),)
    path = os.path.abspath(os.path.join(dirpath, prefix))
    if input_names:
        path = f"{path}-{'-'.join(input_names)}"
    if shape:
        path = f"{path}-{'x'.join(str(s) for s in shape)}"
    if output_names:
        path = f"{path}-{'-'.join(output_names)}"
    return f"{path}.json"

# Internal Cell
class _PartialDense:
    def __init__(self, weights, x_norm, y_norm, input_names, output_names):
        self.weights = weights
        self.x_norm = x_norm
        self.y_norm = y_norm
        self.input_names = input_names
        self.output_names = output_names

    def __call__(self, *params: ComplexFloat) -> ComplexFloat:
        return dense(self.weights, *params, x_norm=self.x_norm, y_norm=self.y_norm)

    def __repr__(self):
        return f"{self.__class__.__name__}{repr(self.input_names)}->{repr(self.output_names)}"

# Cell
def load_nn_dense(
    *sizes: int,
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    weightprefix="dense",
    weightdirpath="weights",
    normdirpath="norms",
    normprefix="norm",
    preprocess=preprocess,
) -> Callable:
    """Load a pre-trained dense model"""
    weights_path = get_dense_weights_path(
        *sizes,
        input_names=input_names,
        output_names=output_names,
        prefix=weightprefix,
        dirpath=weightdirpath,
        preprocess=preprocess,
    )
    if not os.path.exists(weights_path):
        raise ValueError("Cannot find weights path for given parameters")
    x_norm_path = get_norm_path(
        input_names=input_names,
        prefix=normprefix,
        dirpath=normdirpath,
        preprocess=preprocess,
    )
    if not os.path.exists(x_norm_path):
        raise ValueError("Cannot find normalization for input parameters")
    y_norm_path = get_norm_path(
        output_names=output_names,
        prefix=normprefix,
        dirpath=normdirpath,
        preprocess=preprocess,
    )
    if not os.path.exists(x_norm_path):
        raise ValueError("Cannot find normalization for output parameters")
    weights = load_nn_weights_json(weights_path)
    x_norm_dict = load_nn_weights_json(x_norm_path)
    y_norm_dict = load_nn_weights_json(y_norm_path)
    x_norm = norm(x_norm_dict["mean"], x_norm_dict["std"])
    y_norm = norm(y_norm_dict["mean"], y_norm_dict["std"])
    partial_dense = _PartialDense(weights, x_norm, y_norm, input_names, output_names)
    return partial_dense


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/08c_backends_additive.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['split_port', 'graph_edges', 'prune_internal_output_nodes', 'get_possible_paths', 'path_lengths',
           'evaluate_circuit_additive']

# Cell
#nbdev_comment from __future__ import annotations

from typing import Dict, Tuple

import networkx as nx
from ..typing_ import SDict, SType, sdict

try:
    import jax.numpy as jnp
    JAX_AVAILABLE = True
except ImportError:
    import numpy as jnp
    JAX_AVAILABLE = False

# Cell
def split_port(port: str) -> Tuple[str, str]:
    try:
        instance, port = port.split(",")
    except ValueError:
        (port,) = port.split(",")
        instance = ""
    return instance, port

# Cell
def graph_edges(
    instances: Dict[str, SType], connections: Dict[str, str], ports: Dict[str, str]
):
    zero = jnp.array([0.0], dtype=float)
    edges = {}
    edges.update({split_port(k): split_port(v) for k, v in connections.items()})
    edges.update({split_port(v): split_port(k) for k, v in connections.items()})
    edges.update({split_port(k): split_port(v) for k, v in ports.items()})
    edges.update({split_port(v): split_port(k) for k, v in ports.items()})
    edges = [(n1, n2, {"type": "C", "length": zero}) for n1, n2 in edges.items()]

    _instances = {
        **{i1: None for (i1, _), (_, _), _ in edges},
        **{i2: None for (_, _), (i2, _), _ in edges},
    }
    del _instances[""] # external ports don't belong to an instance

    for instance in _instances:
        s = instances[instance]
        edges += [
            ((instance, p1), (instance, p2), {"type": "S", "length": jnp.asarray(length, dtype=float).ravel()})
            for (p1, p2), length in sdict(s).items()
        ]

    return edges

# Cell
def prune_internal_output_nodes(graph):
    broken = True
    while broken:
        broken = False
        for (i, p), dic in list(graph.adjacency()):
            if (
                i != ""
                and len(dic) == 2
                and all(prop.get("type", "C") == "C" for prop in dic.values())
            ):
                graph.remove_node((i, p))
                graph.add_edge(*dic.keys(), type="C", length=0.0)
                broken = True
                break
    return graph

# Cell
def get_possible_paths(graph, source, target):
    paths = []
    default_props = {"type": "C", "length": 0.0}
    for path in nx.all_simple_edge_paths(graph, source, target):
        prevtype = "C"
        for n1, n2 in path:
            curtype = graph.get_edge_data(n1, n2, default_props)["type"]
            if curtype == prevtype == "S":
                break
            else:
                prevtype = curtype
        else:
            paths.append(path)
    return paths

# Cell
def path_lengths(graph, paths):
    lengths = []
    for path in paths:
        length = zero = jnp.array([0.0], dtype=float)
        default_edge_data = {"type": "C", "length": zero}
        for edge in path:
            edge_data = graph.get_edge_data(*edge, default_edge_data)
            length = (length[None,:] + edge_data.get("length", zero)[:, None]).ravel()
        lengths.append(length)
    return lengths

# Cell
def evaluate_circuit_additive(
    instances: Dict[str, SDict],
    connections: Dict[str, str],
    ports: Dict[str, str],
):
    """evaluate a circuit for the given sdicts."""
    edges = graph_edges(instances, connections, ports)

    graph = nx.Graph()
    graph.add_edges_from(edges)
    prune_internal_output_nodes(graph)

    sdict = {}
    for source in ports:
        for target in ports:
            paths = get_possible_paths(graph, source=("", source), target=("", target))
            if not paths:
                continue
            sdict[source, target] = path_lengths(graph, paths)

    return sdict


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_circuit.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['create_dag', 'draw_dag', 'find_root', 'find_leaves', 'circuit', 'NetlistDict', 'CircuitInfo',
           'RecursiveNetlistDict', 'get_required_circuit_models']

# Cell
#nbdev_comment from __future__ import annotations

import os
import shutil
import sys
from functools import partial
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, TypedDict, Union

import black
import networkx as nx
import numpy as np
from pydantic import ValidationError
from sax import reciprocal
from .backends import circuit_backends
from .multimode import multimode, singlemode
from .netlist import Netlist, RecursiveNetlist
from .netlist_cleaning import remove_unused_instances
from .typing_ import Model, Settings, SType
from .utils import _replace_kwargs, get_settings, merge_dicts, update_settings

# Cell
def create_dag(
    netlist: RecursiveNetlist,
    models: Optional[Dict[str, Any]] = None,
):
    if models is None:
        models = {}
    assert isinstance(models, dict)

    all_models = {}
    g = nx.DiGraph()

    for model_name, subnetlist in netlist.dict().items():
        if not model_name in all_models:
            all_models[model_name] = models.get(model_name, subnetlist)
            g.add_node(model_name)
        if model_name in models:
            continue
        for instance in subnetlist["instances"].values():
            component = instance["component"]
            if not component in all_models:
                all_models[component] = models.get(component, None)
                g.add_node(component)
            g.add_edge(model_name, component)

    # we only need the nodes that depend on the parent...
    parent_node = next(iter(netlist.root.keys()))
    nodes = [parent_node, *nx.descendants(g, parent_node)]
    g = nx.induced_subgraph(g, nodes)

    return g

# Cell


def draw_dag(dag, with_labels=True, **kwargs):
    _patch_path()
    if shutil.which("dot"):
        return nx.draw(
            dag,
            nx.nx_pydot.pydot_layout(dag, prog="dot"),
            with_labels=with_labels,
            **kwargs
        )
    else:
        return nx.draw(dag, _my_dag_pos(dag), with_labels=with_labels, **kwargs)


def _patch_path():
    os_paths = {p: None for p in os.environ.get("PATH", "").split(os.pathsep)}
    sys_paths = {p: None for p in sys.path}
    other_paths = {os.path.dirname(sys.executable): None}
    os.environ["PATH"] = os.pathsep.join({**os_paths, **sys_paths, **other_paths})


def _my_dag_pos(dag):
    # inferior to pydot
    in_degree = {}
    for k, v in dag.in_degree():
        if v not in in_degree:
            in_degree[v] = []
        in_degree[v].append(k)

    widths = {k: len(vs) for k, vs in in_degree.items()}
    width = max(widths.values())
    height = max(widths) + 1

    horizontal_pos = {
        k: np.linspace(0, 1, w + 2)[1:-1] * width for k, w in widths.items()
    }

    pos = {}
    for k, vs in in_degree.items():
        for x, v in zip(horizontal_pos[k], vs):
            pos[v] = (x, -k)
    return pos

# Cell
def find_root(g):
    nodes = [n for n, d in g.in_degree() if d == 0]
    return nodes

# Cell
def find_leaves(g):
    nodes = [n for n, d in g.out_degree() if d == 0]
    return nodes

# Cell
def _validate_models(models, dag):
    required_models = find_leaves(dag)
    missing_models = [m for m in required_models if m not in models]
    if missing_models:
        model_diff = {
            "Missing Models": missing_models,
            "Given Models": list(models),
            "Required Models": required_models,
        }
        raise ValueError(
            "Missing models. The following models are still missing to build the circuit:\n"
            f"{black.format_str(repr(model_diff), mode=black.Mode())}"
        )
    return {**models} # shallow copy

# Cell
def _flat_circuit(instances, connections, ports, models, backend):
    evaluate_circuit = circuit_backends[backend]

    inst2model = {k: models[inst.component] for k, inst in instances.items()}

    model_settings = {name: get_settings(model) for name, model in inst2model.items()}
    netlist_settings = {
        name: {k: v for k, v in (inst.settings or {}).items() if k in model_settings[name]}
        for name, inst in instances.items()
    }
    default_settings = merge_dicts(model_settings, netlist_settings)

    def _circuit(**settings: Settings) -> SType:
        settings = merge_dicts(default_settings, settings)
        settings = _forward_global_settings(inst2model, settings)

        instances: Dict[str, SType] = {}
        for inst_name, model in inst2model.items():
            instances[inst_name] = model(**settings.get(inst_name, {}))
        #print(f"{instances=}")
        #print(f"{connections=}")
        #print(f"{ports=}")
        S = evaluate_circuit(instances, connections, ports)
        return S

    _replace_kwargs(_circuit, **default_settings)

    return _circuit

def _forward_global_settings(instances, settings):
    global_settings = {}
    for k in list(settings.keys()):
        if k in instances:
            continue
        global_settings[k] = settings.pop(k)
    if global_settings:
        settings = update_settings(settings, **global_settings)
    return settings

# Cell

def circuit(
    netlist: Union[Netlist, NetlistDict, RecursiveNetlist, RecursiveNetlistDict],
    models: Optional[Dict[str, Model]] = None,
    modes: Optional[List[str]] = None,
    backend: str = "default",
) -> Tuple[Model, CircuitInfo]:
    netlist = _ensure_recursive_netlist_dict(netlist)

    # TODO: do the following two steps *after* recursive netlist parsing.
    netlist = remove_unused_instances(netlist)
    netlist, instance_models = _extract_instance_models(netlist)

    recnet: RecursiveNetlist = _validate_net(netlist)
    dependency_dag: nx.DiGraph = _validate_dag(
        create_dag(recnet, models)
    )  # directed acyclic graph
    models = _validate_models({**(models or {}), **instance_models}, dependency_dag)
    modes = _validate_modes(modes)
    backend = _validate_circuit_backend(backend)

    circuit = None
    new_models = {}
    current_models = {}
    model_names = list(nx.topological_sort(dependency_dag))[::-1]
    for model_name in model_names:
        if model_name in models:
            new_models[model_name] = models[model_name]
            continue

        flatnet = recnet.root[model_name]

        connections, ports, new_models = _make_singlemode_or_multimode(
            flatnet, modes, new_models
        )
        current_models.update(new_models)
        new_models = {}

        current_models[model_name] = circuit = _flat_circuit(
            flatnet.instances, connections, ports, current_models, backend
        )

    assert circuit is not None
    return circuit, CircuitInfo(dag=dependency_dag, models=current_models)


class NetlistDict(TypedDict):
    instances: Dict
    connections: Dict[str, str]
    ports: Dict[str, str]


RecursiveNetlistDict = Dict[str, NetlistDict]


class CircuitInfo(NamedTuple):
    dag: nx.DiGraph
    models: Dict[str, Model]


def _ensure_recursive_netlist_dict(netlist):
    if not isinstance(netlist, dict):
        netlist = netlist.dict()
    if "instances" in netlist:
        netlist = {"top_level": netlist}
    netlist = {**netlist}
    for k, v in netlist.items():
        netlist[k] = {**v}
    return netlist


def _extract_instance_models(netlist):
    models = {}
    for netname, net in netlist.items():
        net = {**net}
        net["instances"] = {**net["instances"]}
        for name, inst in net["instances"].items():
            if callable(inst):
                settings = get_settings(inst)
                if isinstance(inst, partial) and inst.args:
                    raise ValueError(
                        "SAX circuits and netlists don't support partials with positional arguments."
                    )
                while isinstance(inst, partial):
                    inst = inst.func
                models[inst.__name__] = inst
                net["instances"][name] = {
                    "component": inst.__name__,
                    "settings": settings,
                }
        netlist[netname] = net
    return netlist, models


def _validate_circuit_backend(backend):
    backend = backend.lower()
    # assert valid circuit_backend
    if backend not in circuit_backends:
        raise KeyError(
            f"circuit backend {backend} not found. Allowed circuit backends: "
            f"{', '.join(circuit_backends.keys())}."
        )
    return backend


def _validate_modes(modes) -> List[str]:
    if modes is None:
        return ["te"]
    elif not modes:
        return ["te"]
    elif isinstance(modes, str):
        return [modes]
    elif all(isinstance(m, str) for m in modes):
        return modes
    else:
        raise ValueError(f"Invalid modes given: {modes}")


def _validate_net(netlist: Union[Netlist, RecursiveNetlist]) -> RecursiveNetlist:
    if isinstance(netlist, dict):
        try:
            netlist = Netlist.model_validate(netlist)
        except ValidationError:
            netlist = RecursiveNetlist.model_validate(netlist)
    elif isinstance(netlist, Netlist):
        netlist = RecursiveNetlist(root={"top_level": netlist})
    return netlist


def _validate_dag(dag):
    nodes = find_root(dag)
    if len(nodes) > 1:
        raise ValueError(f"Multiple top_levels found in netlist: {nodes}")
    if len(nodes) < 1:
        raise ValueError(f"Netlist does not contain any nodes.")
    if not dag.is_directed():
        raise ValueError("Netlist dependency cycles detected!")
    return dag


def _make_singlemode_or_multimode(netlist, modes, models):
    if len(modes) == 1:
        connections, ports, models = _make_singlemode(netlist, modes[0], models)
    else:
        connections, ports, models = _make_multimode(netlist, modes, models)
    return connections, ports, models


def _make_singlemode(netlist, mode, models):
    models = {k: singlemode(m, mode=mode) for k, m in models.items()}
    return netlist.connections, netlist.ports, models


def _make_multimode(netlist, modes, models):
    models = {k: multimode(m, modes=modes) for k, m in models.items()}
    connections = {
        f"{p1}@{mode}": f"{p2}@{mode}"
        for p1, p2 in netlist.connections.items()
        for mode in modes
    }
    ports = {
        f"{p1}@{mode}": f"{p2}@{mode}"
        for p1, p2 in netlist.ports.items()
        for mode in modes
    }
    return connections, ports, models

# Cell
def get_required_circuit_models(
        netlist: Union[Netlist, NetlistDict, RecursiveNetlist, RecursiveNetlistDict],
        models: Optional[Dict[str, Model]] = None,
) -> List:
    if models is None:
        models = {}
    assert isinstance(models, dict)
    netlist = _ensure_recursive_netlist_dict(netlist)
    # TODO: do the following two steps *after* recursive netlist parsing.
    netlist = remove_unused_instances(netlist)
    netlist, instance_models = _extract_instance_models(netlist)
    recnet: RecursiveNetlist = _validate_net(netlist)

    missing_models = {}
    missing_model_names = []
    g = nx.DiGraph()

    for model_name, subnetlist in recnet.dict().items():
        if not model_name in missing_models:
            missing_models[model_name] = models.get(model_name, subnetlist)
            g.add_node(model_name)
        if model_name in models:
            continue
        for instance in subnetlist["instances"].values():
            component = instance["component"]
            if (not component in missing_models) and (not component in models):
                missing_models[component] = models.get(component, None)
                missing_model_names.append(component)
                g.add_node(component)
            g.add_edge(model_name, component)
    return missing_model_names


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_models.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['straight', 'coupler', 'unitary', 'passthru', 'copier', 'passthru', 'get_models', 'models']

# Cell
#nbdev_comment from __future__ import annotations

import warnings
from functools import lru_cache as cache
from types import SimpleNamespace
from typing import Optional, Tuple

import sax
from .typing_ import Model, SCoo, SDict
from .utils import get_inputs_outputs, reciprocal

try:
    import jax
    import jax.numpy as jnp
    JAX_AVAILABLE = True
except ImportError:
    import numpy as jnp

    def jit(func, *args, **kwargs):
        warnings.warn("[NO JAX] skipping jit! Please install JAX!")
        return func

    jax = SimpleNamespace(jit=jit)

    JAX_AVAILABLE = False

# Cell

def straight(
    *,
    wl: float = 1.55,
    wl0: float = 1.55,
    neff: float = 2.34,
    ng: float = 3.4,
    length: float = 10.0,
    loss: float = 0.0
) -> SDict:
    """a simple straight waveguide model"""
    dwl = wl - wl0
    dneff_dwl = (ng - neff) / wl0
    neff = neff - dwl * dneff_dwl
    phase = 2 * jnp.pi * neff * length / wl
    amplitude = jnp.asarray(10 ** (-loss * length / 20), dtype=complex)
    transmission =  amplitude * jnp.exp(1j * phase)
    sdict = reciprocal(
        {
            ("in0", "out0"): transmission,
        }
    )
    return sdict

# Cell

def coupler(*, coupling: float = 0.5) -> SDict:
    """a simple coupler model"""
    kappa = coupling ** 0.5
    tau = (1 - coupling) ** 0.5
    sdict = reciprocal(
        {
            ("in0", "out0"): tau,
            ("in0", "out1"): 1j * kappa,
            ("in1", "out0"): 1j * kappa,
            ("in1", "out1"): tau,
        }
    )
    return sdict

# Internal Cell

def _validate_ports(ports, num_inputs, num_outputs, diagonal) -> Tuple[Tuple[str,...], Tuple[str,...], int, int]:
    if ports is None:
        if num_inputs is None or num_outputs is None:
            raise ValueError(
                "if not ports given, you must specify how many input ports "
                "and how many output ports a model has."
            )
        input_ports = [f"in{i}" for i in range(num_inputs)]
        output_ports = [f"out{i}" for i in range(num_outputs)]
    else:
        if num_inputs is not None:
            if num_outputs is None:
                raise ValueError(
                    "if num_inputs is given, num_outputs should be given as well."
                )
        if num_outputs is not None:
            if num_inputs is None:
                raise ValueError(
                    "if num_outputs is given, num_inputs should be given as well."
                )
        if num_inputs is not None and num_outputs is not None:
            if num_inputs + num_outputs != len(ports):
                raise ValueError("num_inputs + num_outputs != len(ports)")
            input_ports = ports[:num_inputs]
            output_ports = ports[num_inputs:]
        else:
            input_ports, output_ports = get_inputs_outputs(ports)
            num_inputs = len(input_ports)
            num_outputs = len(output_ports)

    if diagonal:
        if num_inputs != num_outputs:
            raise ValueError(
                "Can only have a diagonal passthru if number of input ports equals the number of output ports!"
            )
    return input_ports, output_ports, num_inputs, num_outputs

# Cell

@cache
def unitary(
    num_inputs: Optional[int] = None,
    num_outputs: Optional[int] = None,
    ports: Optional[Tuple[str, ...]] = None,
    *,
    jit=True,
    reciprocal=True,
    diagonal=False,
) -> Model:
    input_ports, output_ports, num_inputs, num_outputs = _validate_ports(ports, num_inputs, num_outputs, diagonal)
    assert num_inputs is not None and num_outputs is not None

    # let's create the squared S-matrix:
    N = max(num_inputs, num_outputs)
    S = jnp.zeros((2*N, 2*N), dtype=float)

    if not diagonal:
        if JAX_AVAILABLE:
            S = S.at[:N, N:].set(1)
        else:
            S[:N, N:] = 1
    else:
        r = jnp.arange(N, dtype=int) # reciprocal only works if num_inputs == num_outputs!
        if JAX_AVAILABLE:
            S = S.at[r, N+r].set(1)
        else:
            S[r, N+r] = 1

    if reciprocal:
        if not diagonal:
            if JAX_AVAILABLE:
                S = S.at[N:, :N].set(1)
            else:
                S[N:, :N] = 1
        else:
            r = jnp.arange(N, dtype=int) # reciprocal only works if num_inputs == num_outputs!
            if JAX_AVAILABLE:
                S = S.at[N+r, r].set(1)
            else:
                S[N+r, r] = 1

    # Now we need to normalize the squared S-matrix
    U, s, V = jnp.linalg.svd(S, full_matrices=False)
    S = jnp.sqrt(U@jnp.diag(jnp.where(s > 1e-12, 1, 0))@V)

    # Now create subset of this matrix we're interested in:
    r = jnp.concatenate([jnp.arange(num_inputs, dtype=int), N+jnp.arange(num_outputs, dtype=int)], 0)
    S = S[r, :][:, r]

    # let's convert it in SCOO format:
    Si, Sj = jnp.where(S > 1e-6)
    Sx = S[Si, Sj]

    # the last missing piece is a port map:
    pm = {
        **{p: i for i, p in enumerate(input_ports)},
        **{p: i + num_inputs for i, p in enumerate(output_ports)},
    }

    def func(wl: float = 1.5) -> SCoo:
        wl_ = jnp.asarray(wl)
        Sx_ = jnp.broadcast_to(Sx, (*wl_.shape, *Sx.shape))
        return Si, Sj, Sx_, pm

    func.__name__ = f"unitary_{num_inputs}_{num_outputs}"
    func.__qualname__ = f"unitary_{num_inputs}_{num_outputs}"
    if jit:
        return jax.jit(func)
    return func

# Cell
@cache
def passthru(
    num_links: Optional[int] = None,
    ports: Optional[Tuple[str, ...]] = None,
    *,
    jit=True,
    reciprocal=True,
) -> Model:
    passthru = unitary(num_links, num_links, ports, jit=jit, reciprocal=reciprocal, diagonal=True)
    passthru.__name__ = f"passthru_{num_links}_{num_links}"
    passthru.__qualname__ = f"passthru_{num_links}_{num_links}"
    if jit:
        return jax.jit(passthru)
    return passthru

# Cell

@cache
def copier(
    num_inputs: Optional[int] = None,
    num_outputs: Optional[int] = None,
    ports: Optional[Tuple[str, ...]] = None,
    *,
    jit=True,
    reciprocal=True,
    diagonal=False,
) -> Model:

    input_ports, output_ports, num_inputs, num_outputs = _validate_ports(ports, num_inputs, num_outputs, diagonal)
    assert num_inputs is not None and num_outputs is not None

    # let's create the squared S-matrix:
    S = jnp.zeros((num_inputs+num_outputs, num_inputs+num_outputs), dtype=float)

    if not diagonal:
        if JAX_AVAILABLE:
            S = S.at[:num_inputs, num_inputs:].set(1)
        else:
            S[:num_inputs, num_inputs:] = 1
    else:
        r = jnp.arange(num_inputs, dtype=int) # == range(num_outputs) # reciprocal only works if num_inputs == num_outputs!
        if JAX_AVAILABLE:
            S = S.at[r, num_inputs+r].set(1)
        else:
            S[r, num_inputs+r] = 1

    if reciprocal:
        if not diagonal:
            if JAX_AVAILABLE:
                S = S.at[num_inputs:, :num_inputs].set(1)
            else:
                S[num_inputs:, :num_inputs] = 1
        else:
            r = jnp.arange(num_inputs, dtype=int) # == range(num_outputs) # reciprocal only works if num_inputs == num_outputs!
            if JAX_AVAILABLE:
                S = S.at[num_inputs+r, r].set(1)
            else:
                S[num_inputs+r, r] = 1

    # let's convert it in SCOO format:
    Si, Sj = jnp.where(S > 1e-6)
    Sx = S[Si, Sj]

    # the last missing piece is a port map:
    pm = {
        **{p: i for i, p in enumerate(input_ports)},
        **{p: i + num_inputs for i, p in enumerate(output_ports)},
    }

    def func(wl: float = 1.5) -> SCoo:
        wl_ = jnp.asarray(wl)
        Sx_ = jnp.broadcast_to(Sx, (*wl_.shape, *Sx.shape))
        return Si, Sj, Sx_, pm

    func.__name__ = f"unitary_{num_inputs}_{num_outputs}"
    func.__qualname__ = f"unitary_{num_inputs}_{num_outputs}"
    if jit:
        return jax.jit(func)
    return func

# Cell
@cache
def passthru(
    num_links: Optional[int] = None,
    ports: Optional[Tuple[str, ...]] = None,
    *,
    jit=True,
    reciprocal=True,
) -> Model:
    passthru = unitary(num_links, num_links, ports, jit=jit, reciprocal=reciprocal, diagonal=True)
    passthru.__name__ = f"passthru_{num_links}_{num_links}"
    passthru.__qualname__ = f"passthru_{num_links}_{num_links}"
    if jit:
        return jax.jit(passthru)
    return passthru

# Cell

models = {
    "copier": copier,
    "coupler": coupler,
    "passthru": passthru,
    "straight": straight,
    "unitary": unitary,
}

def get_models(copy: bool=True):
    if copy:
        return {**models}
    return models


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/09_nn.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = []

# Cell
#nbdev_comment from __future__ import annotations

# Cell

from .loss import huber_loss as huber_loss
from .loss import l2_reg as l2_reg
from .loss import mse as mse

# Cell

from .utils import (
    cartesian_product as cartesian_product,
    denormalize as denormalize,
    get_normalization as get_normalization,
    get_df_columns as get_df_columns,
    normalize as normalize,
)

# Cell

from .core import (
    preprocess as preprocess,
    dense as dense,
    generate_dense_weights as generate_dense_weights,
)

# Cell

from .io import (
    load_nn_weights_json as load_nn_weights_json,
    save_nn_weights_json as save_nn_weights_json,
    get_available_sizes as get_available_sizes,
    get_dense_weights_path as get_dense_weights_path,
    get_norm_path as get_norm_path,
    load_nn_dense as load_nn_dense,
)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/08b_backends_klu.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['solve_klu', 'mul_coo', 'evaluate_circuit_klu']

# Cell
#nbdev_comment from __future__ import annotations

from typing import Dict


from ..typing_ import SDense, SDict, SType, scoo
from . import evaluate_circuit

try:
    import klujax
except ImportError:
    klujax = None

try:
    import jax
    import jax.numpy as jnp
    JAX_AVAILABLE = True
except ImportError:
    import numpy as jnp
    JAX_AVAILABLE = False

# Cell
solve_klu = None
if klujax is not None:
    solve_klu = jax.vmap(klujax.solve, (None, None, 0, None), 0)

# Internal Cell

# @jax.jit  # TODO: make this available to autograd
# def mul_coo(Ai, Aj, Ax, b):
#     result = jnp.zeros_like(b).at[..., Ai, :].add(Ax[..., :, None] * b[..., Aj, :])
#     return result

# Cell
mul_coo = None
if klujax is not None:
    mul_coo = jax.vmap(klujax.coo_mul_vec, (None, None, 0, 0), 0)

# Cell
def evaluate_circuit_klu(
    instances: Dict[str, SType],
    connections: Dict[str, str],
    ports: Dict[str, str],
):
    """evaluate a circuit using KLU for the given sdicts. """

    if klujax is None:
        raise ImportError(
            "Could not import 'klujax'. "
            "Please install it first before using backend method 'klu'"
        )

    assert solve_klu is not None
    assert mul_coo is not None

    connections = {**connections, **{v: k for k, v in connections.items()}}
    inverse_ports = {v: k for k, v in ports.items()}
    port_map = {k: i for i, k in enumerate(ports)}

    idx, Si, Sj, Sx, instance_ports = 0, [], [], [], {}
    batch_shape = ()
    for name, instance in instances.items():
        si, sj, sx, ports_map = scoo(instance)
        Si.append(si + idx)
        Sj.append(sj + idx)
        Sx.append(sx)
        if len(sx.shape[:-1]) > len(batch_shape):
            batch_shape = sx.shape[:-1]
        instance_ports.update({f"{name},{p}": i + idx for p, i in ports_map.items()})
        idx += len(ports_map)

    Si = jnp.concatenate(Si, -1)
    Sj = jnp.concatenate(Sj, -1)
    Sx = jnp.concatenate(
        [jnp.broadcast_to(sx, (*batch_shape, sx.shape[-1])) for sx in Sx], -1
    )

    n_col = idx
    n_rhs = len(port_map)

    Cmap = {
        int(instance_ports[k]): int(instance_ports[v]) for k, v in connections.items()
    }
    Ci = jnp.array(list(Cmap.keys()), dtype=jnp.int32)
    Cj = jnp.array(list(Cmap.values()), dtype=jnp.int32)

    Cextmap = {int(instance_ports[k]): int(port_map[v]) for k, v in inverse_ports.items()}
    Cexti = jnp.stack(list(Cextmap.keys()), 0)
    Cextj = jnp.stack(list(Cextmap.values()), 0)
    Cext = jnp.zeros((n_col, n_rhs), dtype=complex).at[Cexti, Cextj].set(1.0)

    # TODO: make this block jittable...
    Ix = jnp.ones((*batch_shape, n_col))
    Ii = Ij = jnp.arange(n_col)
    mask = Cj[None,:] == Si[:, None]
    CSi = jnp.broadcast_to(Ci[None, :], mask.shape)[mask]

    # CSi = jnp.where(Cj[None, :] == Si[:, None], Ci[None, :], 0).sum(1)
    mask = (Cj[:, None] == Si[None, :]).any(0)
    CSj = Sj[mask]

    if Sx.ndim > 1: # bug in JAX... see https://github.com/google/jax/issues/9050
        CSx = Sx[..., mask]
    else:
        CSx = Sx[mask]

    # CSj = jnp.where(mask, Sj, 0)
    # CSx = jnp.where(mask, Sx, 0.0)

    I_CSi = jnp.concatenate([CSi, Ii], -1)
    I_CSj = jnp.concatenate([CSj, Ij], -1)
    I_CSx = jnp.concatenate([-CSx, Ix], -1)

    n_col, n_rhs = Cext.shape
    n_lhs = jnp.prod(jnp.array(batch_shape, dtype=jnp.int32))
    Sx = Sx.reshape(n_lhs, -1)
    I_CSx = I_CSx.reshape(n_lhs, -1)

    inv_I_CS_Cext = solve_klu(I_CSi, I_CSj, I_CSx, Cext)
    S_inv_I_CS_Cext = mul_coo(Si, Sj, Sx, inv_I_CS_Cext)

    CextT_S_inv_I_CS_Cext = S_inv_I_CS_Cext[..., Cexti, :][..., :, Cextj]

    _, n, _ = CextT_S_inv_I_CS_Cext.shape
    S = CextT_S_inv_I_CS_Cext.reshape(*batch_shape, n, n)

    return S, port_map


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/09b_nn_utils.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['cartesian_product', 'denormalize', 'get_normalization', 'get_df_columns', 'normalize']

# Cell
#nbdev_comment from __future__ import annotations

from collections import namedtuple
from typing import Tuple

import jax.numpy as jnp
import pandas as pd
from ..typing_ import ComplexFloat

# Cell
def cartesian_product(*arrays: ComplexFloat) -> ComplexFloat:
    """calculate the n-dimensional cartesian product of an arbitrary number of arrays"""
    ixarrays = jnp.ix_(*arrays)
    barrays = jnp.broadcast_arrays(*ixarrays)
    sarrays = jnp.stack(barrays, -1)
    assert isinstance(sarrays, jnp.ndarray)
    product = sarrays.reshape(-1, sarrays.shape[-1])
    assert isinstance(product, jnp.ndarray)
    return product

# Cell
def denormalize(x: ComplexFloat, mean: ComplexFloat = 0.0, std: ComplexFloat = 1.0) -> ComplexFloat:
    """denormalize an array with a given mean and standard deviation"""
    return x * std + mean

# Internal Cell
norm = namedtuple("norm", ("mean", "std"))

# Cell
def get_normalization(x: ComplexFloat):
    """Get mean and standard deviation for a given array"""
    if isinstance(x, (complex, float)):
        return x, 0.0
    return norm(x.mean(0), x.std(0))

# Cell
def get_df_columns(df: pd.DataFrame, *names: str) -> Tuple[ComplexFloat, ...]:
    """Get certain columns from a pandas DataFrame as jax.numpy arrays"""
    tup = namedtuple("params", names)
    params_list = []
    for name in names:
        column_np = df[name].values
        column_jnp = jnp.array(column_np)
        assert isinstance(column_jnp, jnp.ndarray)
        params_list.append(column_jnp.ravel())
    return tup(*params_list)

# Cell
def normalize(
    x: ComplexFloat, mean: ComplexFloat = 0.0, std: ComplexFloat = 1.0
) -> ComplexFloat:
    """normalize an array with a given mean and standard deviation"""
    return (x - mean) / std


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_typing.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['Array', 'Int', 'Float', 'ComplexFloat', 'Settings', 'SDict', 'SCoo', 'SDense', 'SType', 'Model',
           'ModelFactory', 'Models', 'is_float', 'is_complex', 'is_complex_float', 'is_sdict', 'is_scoo', 'is_sdense',
           'is_model', 'is_model_factory', 'validate_model', 'is_stype', 'is_singlemode', 'is_multimode',
           'is_mixedmode', 'sdict', 'scoo', 'sdense', 'modelfactory']

# Cell
#nbdev_comment from __future__ import annotations
import functools
import inspect
from collections.abc import Callable as CallableABC
from typing import Any, Callable, Dict, Tuple, Union, cast, overload
try:
    from typing import TypedDict
except ImportError: # python<3.8
    from typing_extensions import TypedDict

import numpy as np
from natsort import natsorted

try:
    import jax.numpy as jnp
    JAX_AVAILABLE = True
except ImportError:
    import numpy as jnp
    JAX_AVAILABLE = False

# Cell
Array = Union[jnp.ndarray, np.ndarray]

# Cell
Int = Union[int, Array]

# Cell
Float = Union[float, Array]

# Cell
ComplexFloat = Union[complex, Float]

# Cell
Settings = Union[Dict[str, ComplexFloat], Dict[str, "Settings"]]

# Cell
SDict = Dict[Tuple[str, str], ComplexFloat]

# Cell
SCoo = Tuple[Array, Array, ComplexFloat, Dict[str, int]]

# Cell
SDense = Tuple[Array, Dict[str, int]]

# Cell
SType = Union[SDict, SCoo, SDense]

# Cell
Model = Callable[..., SType]

# Cell
ModelFactory = Callable[..., Model]

# Cell
Models = Dict[str, Model]

# Cell
def is_float(x: Any) -> bool:
    """Check if an object is a `Float`"""
    if isinstance(x, float):
        return True
    if isinstance(x, np.ndarray):
        return x.dtype in (np.float16, np.float32, np.float64, np.float128)
    if isinstance(x, jnp.ndarray):
        return x.dtype in (jnp.float16, jnp.float32, jnp.float64)
    return False

# Cell
def is_complex(x: Any) -> bool:
    """check if an object is a `ComplexFloat`"""
    if isinstance(x, complex):
        return True
    if isinstance(x, np.ndarray):
        return x.dtype in (np.complex64, np.complex128)
    if isinstance(x, jnp.ndarray):
        return x.dtype in (jnp.complex64, jnp.complex128)
    return False

# Cell
def is_complex_float(x: Any) -> bool:
    """check if an object is either a `ComplexFloat` or a `Float`"""
    return is_float(x) or is_complex(x)

# Cell
def is_sdict(x: Any) -> bool:
    """check if an object is an `SDict` (a SAX S-dictionary)"""
    return isinstance(x, dict)

# Cell
def is_scoo(x: Any) -> bool:
    """check if an object is an `SCoo` (a SAX sparse S-matrix representation in COO-format)"""
    return isinstance(x, (tuple, list)) and len(x) == 4

# Cell
def is_sdense(x: Any) -> bool:
    """check if an object is an `SDense` (a SAX dense S-matrix representation)"""
    return isinstance(x, (tuple, list)) and len(x) == 2

# Cell
def is_model(model: Any) -> bool:
    """check if a callable is a `Model` (a callable returning an `SType`)"""
    if not callable(model):
        return False
    try:
        sig = inspect.signature(model)
    except ValueError:
        return False
    for param in sig.parameters.values():
        if param.default is inspect.Parameter.empty:
            return False  # a proper SAX model does not have any positional arguments.
    if _is_callable_annotation(sig.return_annotation):  # model factory
        return False
    return True

def _is_callable_annotation(annotation: Any) -> bool:
    """check if an annotation is `Callable`-like"""
    if isinstance(annotation, str):
        # happens when
        # was imported at the top of the file...
        return annotation.startswith("Callable") or annotation.endswith("Model")
        # TODO: this is not a very robust check...
    try:
        return annotation.__origin__ == CallableABC
    except AttributeError:
        return False

# Cell
def is_model_factory(model: Any) -> bool:
    """check if a callable is a model function."""
    if not callable(model):
        return False
    sig = inspect.signature(model)
    if _is_callable_annotation(sig.return_annotation):  # model factory
        return True
    return False

# Cell
def validate_model(model: Callable):
    """Validate the parameters of a model"""
    positional_arguments = []
    for param in inspect.signature(model).parameters.values():
        if param.default is inspect.Parameter.empty:
            positional_arguments.append(param.name)
    if positional_arguments:
        raise ValueError(
            f"model '{model}' takes positional arguments {', '.join(positional_arguments)} "
            "and hence is not a valid SAX Model! A SAX model should ONLY take keyword arguments (or no arguments at all)."
        )

# Cell
def is_stype(stype: Any) -> bool:
    """check if an object is an SDict, SCoo or SDense"""
    return is_sdict(stype) or is_scoo(stype) or is_sdense(stype)

# Cell
def is_singlemode(S: Any) -> bool:
    """check if an stype is single mode"""
    if not is_stype(S):
        return False
    ports = _get_ports(S)
    return not any(("@" in p) for p in ports)

def _get_ports(S: SType):
    if is_sdict(S):
        S = cast(SDict, S)
        ports_set = {p1 for p1, _ in S} | {p2 for _, p2 in S}
        return tuple(natsorted(ports_set))
    else:
        *_, ports_map = S
        assert isinstance(ports_map, dict)
        return tuple(natsorted(ports_map.keys()))

# Cell
def is_multimode(S: Any) -> bool:
    """check if an stype is single mode"""
    if not is_stype(S):
        return False

    ports = _get_ports(S)
    return all(("@" in p) for p in ports)

# Cell
def is_mixedmode(S: Any) -> bool:
    """check if an stype is neither single mode nor multimode (hence invalid)"""
    return not is_singlemode(S) and not is_multimode(S)

# Internal Cell

@overload
def sdict(S: Model) -> Model:
    ...


@overload
def sdict(S: SType) -> SDict:
    ...

# Cell
def sdict(S: Union[Model, SType]) -> Union[Model, SType]:
    """Convert an `SCoo` or `SDense` to `SDict`"""

    if is_model(S):
        model = cast(Model, S)

        @functools.wraps(model)
        def wrapper(**kwargs):
            return sdict(model(**kwargs))

        return wrapper

    elif is_scoo(S):
        x_dict = _scoo_to_sdict(*cast(SCoo, S))
    elif is_sdense(S):
        x_dict = _sdense_to_sdict(*cast(SDense, S))
    elif is_sdict(S):
        x_dict = cast(SDict, S)
    else:
        raise ValueError("Could not convert arguments to sdict.")

    return x_dict


def _scoo_to_sdict(Si: Array, Sj: Array, Sx: Array, ports_map: Dict[str, int]) -> SDict:
    sdict = {}
    inverse_ports_map = {int(i): p for p, i in ports_map.items()}
    for i, (si, sj) in enumerate(zip(Si, Sj)):
        sdict[
            inverse_ports_map.get(int(si), ""), inverse_ports_map.get(int(sj), "")
        ] = Sx[..., i]
    sdict = {(p1, p2): v for (p1, p2), v in sdict.items() if p1 and p2}
    return sdict


def _sdense_to_sdict(S: Array, ports_map: Dict[str, int]) -> SDict:
    sdict = {}
    for p1, i in ports_map.items():
        for p2, j in ports_map.items():
            sdict[p1, p2] = S[..., i, j]
    return sdict

# Internal Cell

@overload
def scoo(S: Callable) -> Callable:
    ...


@overload
def scoo(S: SType) -> SCoo:
    ...

# Cell

def scoo(S: Union[Callable, SType]) -> Union[Callable, SCoo]:
    """Convert an `SDict` or `SDense` to `SCoo`"""

    if is_model(S):
        model = cast(Model, S)

        @functools.wraps(model)
        def wrapper(**kwargs):
            return scoo(model(**kwargs))

        return wrapper

    elif is_scoo(S):
        S = cast(SCoo, S)
    elif is_sdense(S):
        S = _sdense_to_scoo(*cast(SDense, S))
    elif is_sdict(S):
        S = _sdict_to_scoo(cast(SDict, S))
    else:
        raise ValueError("Could not convert arguments to scoo.")

    return S

def _consolidate_sdense(S, pm):
    idxs = list(pm.values())
    S = S[..., idxs, :][..., :, idxs]
    pm = {p: i for i, p in enumerate(pm)}
    return S, pm

def _sdense_to_scoo(S: Array, ports_map: Dict[str, int]) -> SCoo:
    S, ports_map = _consolidate_sdense(S, ports_map)
    Sj, Si = jnp.meshgrid(jnp.arange(S.shape[-1]), jnp.arange(S.shape[-2]))
    return Si.ravel(), Sj.ravel(), S.reshape(*S.shape[:-2], -1), ports_map


def _sdict_to_scoo(sdict: SDict) -> SCoo:
    all_ports = {}
    for p1, p2 in sdict:
        all_ports[p1] = None
        all_ports[p2] = None
    ports_map = {p: i for i, p in enumerate(all_ports)}
    Sx = jnp.stack(jnp.broadcast_arrays(*sdict.values()), -1)
    Si = jnp.array([ports_map[p] for p, _ in sdict])
    Sj = jnp.array([ports_map[p] for _, p in sdict])
    return Si, Sj, Sx, ports_map

# Internal Cell

@overload
def sdense(S: Callable) -> Callable:
    ...


@overload
def sdense(S: SType) -> SDense:
    ...

# Cell

def sdense(S: Union[Callable, SType]) -> Union[Callable, SDense]:
    """Convert an `SDict` or `SCoo` to `SDense`"""

    if is_model(S):
        model = cast(Model, S)

        @functools.wraps(model)
        def wrapper(**kwargs):
            return sdense(model(**kwargs))

        return wrapper

    if is_sdict(S):
        S = _sdict_to_sdense(cast(SDict, S))
    elif is_scoo(S):
        S = _scoo_to_sdense(*cast(SCoo, S))
    elif is_sdense(S):
        S = cast(SDense, S)
    else:
        raise ValueError("Could not convert arguments to sdense.")

    return S


def _scoo_to_sdense(
    Si: Array, Sj: Array, Sx: Array, ports_map: Dict[str, int]
) -> SDense:
    n_col = len(ports_map)
    S = jnp.zeros((*Sx.shape[:-1], n_col, n_col), dtype=complex)
    if JAX_AVAILABLE:
        S = S.at[..., Si, Sj].add(Sx)
    else:
        S[..., Si, Sj] = Sx
    return S, ports_map

def _sdict_to_sdense(sdict: SDict) -> SDense:
    Si, Sj, Sx, ports_map = _sdict_to_scoo(sdict)
    return _scoo_to_sdense(Si, Sj, Sx, ports_map)

# Cell

def modelfactory(func):
    """Decorator that marks a function as `ModelFactory`"""
    sig = inspect.signature(func)
    if _is_callable_annotation(sig.return_annotation):  # already model factory
        return func
    func.__signature__ = sig.replace(return_annotation=Model)
    return func


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/99_init.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = []

# Internal Cell
#nbdev_comment from __future__ import annotations

__author__ = "Floris Laporte"
__version__ = "0.9.1"

# Cell
from functools import partial as partial
from math import pi as pi

from scipy.constants import c as c

try:
    from flax.core.frozen_dict import FrozenDict as FrozenDict
except ImportError:
    FrozenDict = dict

# Cell

from sax import typing_ as typing
from .typing_ import (
    Array as Array,
    ComplexFloat as ComplexFloat,
    Float as Float,
    Model as Model,
    ModelFactory as ModelFactory,
    Models as Models,
    SCoo as SCoo,
    SDense as SDense,
    SDict as SDict,
    Settings as Settings,
    SType as SType,
    is_complex as is_complex,
    is_complex_float as is_complex_float,
    is_float as is_float,
    is_mixedmode as is_mixedmode,
    is_model as is_model,
    is_model_factory as is_model_factory,
    is_multimode as is_multimode,
    is_scoo as is_scoo,
    is_sdense as is_sdense,
    is_sdict as is_sdict,
    is_singlemode as is_singlemode,
    modelfactory as modelfactory,
    scoo as scoo,
    sdense as sdense,
    sdict as sdict,
    validate_model as validate_model,
)

# Cell

from sax import utils as utils
from .utils import (
    block_diag as block_diag,
    clean_string as clean_string,
    copy_settings as copy_settings,
    flatten_dict as flatten_dict,
    get_inputs_outputs as get_inputs_outputs,
    get_port_combinations as get_port_combinations,
    get_ports as get_ports,
    get_settings as get_settings,
    grouped_interp as grouped_interp,
    merge_dicts as merge_dicts,
    mode_combinations as mode_combinations,
    reciprocal as reciprocal,
    rename_params as rename_params,
    rename_ports as rename_ports,
    try_float as try_float,
    unflatten_dict as unflatten_dict,
    update_settings as update_settings,
    validate_multimode as validate_multimode,
    validate_not_mixedmode as validate_not_mixedmode,
    validate_sdict as validate_sdict,
    validate_settings as validate_settings,
)

# Cell

from sax import multimode as multimode
from .multimode import (
    multimode as multimode,
    singlemode as singlemode,
)

# Cell

from sax import models as models
from .models import get_models as get_models, passthru as passthru

# Cell

from .netlist import netlist as netlist
from .netlist import load_netlist as load_netlist
from .netlist import load_recursive_netlist as load_recursive_netlist
from .netlist import get_netlist_instances_by_prefix as get_netlist_instances_by_prefix
from .netlist import get_component_instances as get_component_instances

# Cell

from .circuit import circuit as circuit
from .circuit import get_required_circuit_models as get_required_circuit_models

# Cell

from sax import backends as backends

# Cell

from sax import patched as _patched


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Array": "00_typing.ipynb",
         "Int": "00_typing.ipynb",
         "Float": "00_typing.ipynb",
         "ComplexFloat": "00_typing.ipynb",
         "Settings": "00_typing.ipynb",
         "SDict": "00_typing.ipynb",
         "SCoo": "00_typing.ipynb",
         "SDense": "00_typing.ipynb",
         "SType": "00_typing.ipynb",
         "Model": "00_typing.ipynb",
         "ModelFactory": "00_typing.ipynb",
         "Models": "00_typing.ipynb",
         "is_float": "00_typing.ipynb",
         "is_complex": "00_typing.ipynb",
         "is_complex_float": "00_typing.ipynb",
         "is_sdict": "00_typing.ipynb",
         "is_scoo": "00_typing.ipynb",
         "is_sdense": "00_typing.ipynb",
         "is_model": "00_typing.ipynb",
         "is_model_factory": "00_typing.ipynb",
         "validate_model": "00_typing.ipynb",
         "is_stype": "00_typing.ipynb",
         "is_singlemode": "00_typing.ipynb",
         "is_multimode": "00_typing.ipynb",
         "is_mixedmode": "00_typing.ipynb",
         "sdict": "00_typing.ipynb",
         "scoo": "00_typing.ipynb",
         "sdense": "00_typing.ipynb",
         "modelfactory": "00_typing.ipynb",
         "block_diag": "02_utils.ipynb",
         "clean_string": "02_utils.ipynb",
         "copy_settings": "02_utils.ipynb",
         "validate_settings": "02_utils.ipynb",
         "try_float": "02_utils.ipynb",
         "flatten_dict": "02_utils.ipynb",
         "unflatten_dict": "02_utils.ipynb",
         "get_ports": "02_utils.ipynb",
         "get_port_combinations": "02_utils.ipynb",
         "get_settings": "02_utils.ipynb",
         "grouped_interp": "02_utils.ipynb",
         "merge_dicts": "02_utils.ipynb",
         "mode_combinations": "02_utils.ipynb",
         "reciprocal": "02_utils.ipynb",
         "rename_params": "02_utils.ipynb",
         "rename_ports": "02_utils.ipynb",
         "update_settings": "02_utils.ipynb",
         "validate_not_mixedmode": "02_utils.ipynb",
         "validate_multimode": "02_utils.ipynb",
         "validate_sdict": "02_utils.ipynb",
         "get_inputs_outputs": "02_utils.ipynb",
         "hash_dict": "02_utils.ipynb",
         "multimode": "04_multimode.ipynb",
         "singlemode": "04_multimode.ipynb",
         "straight": "05_models.ipynb",
         "coupler": "05_models.ipynb",
         "unitary": "05_models.ipynb",
         "passthru": "05_models.ipynb",
         "copier": "05_models.ipynb",
         "get_models": "05_models.ipynb",
         "models": "05_models.ipynb",
         "BaseModel": "06a_netlist.ipynb",
         "Component": "06a_netlist.ipynb",
         "PortEnum": "06a_netlist.ipynb",
         "Placement": "06a_netlist.ipynb",
         "Route": "06a_netlist.ipynb",
         "Netlist": "06a_netlist.ipynb",
         "RecursiveNetlist": "06a_netlist.ipynb",
         "netlist": "06a_netlist.ipynb",
         "load_netlist": "06a_netlist.ipynb",
         "load_recursive_netlist": "06a_netlist.ipynb",
         "get_netlist_instances_by_prefix": "06a_netlist.ipynb",
         "get_component_instances": "06a_netlist.ipynb",
         "remove_unused_instances": "06b_netlist_cleaning.ipynb",
         "create_dag": "07_circuit.ipynb",
         "draw_dag": "07_circuit.ipynb",
         "find_root": "07_circuit.ipynb",
         "find_leaves": "07_circuit.ipynb",
         "circuit": "07_circuit.ipynb",
         "NetlistDict": "07_circuit.ipynb",
         "CircuitInfo": "07_circuit.ipynb",
         "RecursiveNetlistDict": "07_circuit.ipynb",
         "get_required_circuit_models": "07_circuit.ipynb",
         "circuit_backends": "08_backends.ipynb",
         "evaluate_circuit": "08a_backends_default.ipynb",
         "solve_klu": "08b_backends_klu.ipynb",
         "mul_coo": "08b_backends_klu.ipynb",
         "evaluate_circuit_klu": "08b_backends_klu.ipynb",
         "split_port": "08c_backends_additive.ipynb",
         "graph_edges": "08c_backends_additive.ipynb",
         "prune_internal_output_nodes": "08c_backends_additive.ipynb",
         "get_possible_paths": "08c_backends_additive.ipynb",
         "path_lengths": "08c_backends_additive.ipynb",
         "evaluate_circuit_additive": "08c_backends_additive.ipynb",
         "mse": "09a_nn_loss.ipynb",
         "huber_loss": "09a_nn_loss.ipynb",
         "l2_reg": "09a_nn_loss.ipynb",
         "cartesian_product": "09b_nn_utils.ipynb",
         "denormalize": "09b_nn_utils.ipynb",
         "norm": "09b_nn_utils.ipynb",
         "get_normalization": "09b_nn_utils.ipynb",
         "get_df_columns": "09b_nn_utils.ipynb",
         "normalize": "09b_nn_utils.ipynb",
         "preprocess": "09c_nn_core.ipynb",
         "dense": "09c_nn_core.ipynb",
         "generate_dense_weights": "09c_nn_core.ipynb",
         "load_nn_weights_json": "09d_nn_io.ipynb",
         "save_nn_weights_json": "09d_nn_io.ipynb",
         "get_available_sizes": "09d_nn_io.ipynb",
         "get_dense_weights_path": "09d_nn_io.ipynb",
         "get_norm_path": "09d_nn_io.ipynb",
         "load_nn_dense": "09d_nn_io.ipynb",
         "MAGIC_COMMENTS": "98_make_docs.ipynb",
         "load_nb": "98_make_docs.ipynb",
         "repository_path": "98_make_docs.ipynb",
         "docs_path": "98_make_docs.ipynb",
         "save_nb": "98_make_docs.ipynb",
         "strip_metadata": "98_make_docs.ipynb",
         "iter_code_cells": "98_make_docs.ipynb",
         "first_code_cell": "98_make_docs.ipynb",
         "get_default_exp": "98_make_docs.ipynb",
         "iter_function_names": "98_make_docs.ipynb",
         "docs_copy_nb": "98_make_docs.ipynb",
         "list_notebooks": "98_make_docs.ipynb",
         "list_zips": "98_make_docs.ipynb",
         "docs_copy_dir": "98_make_docs.ipynb",
         "docs_copy_simulations": "98_make_docs.ipynb",
         "get_toc_part": "98_make_docs.ipynb",
         "make_docs": "98_make_docs.ipynb",
         "__author__": "99_init.ipynb",
         "__version__": "99_init.ipynb"}

modules = ["typing_.py",
           "patched.py",
           "utils.py",
           "multimode.py",
           "models.py",
           "netlist.py",
           "netlist_cleaning.py",
           "circuit.py",
           "backends/__init__.py",
           "backends/default.py",
           "backends/klu.py",
           "backends/additive.py",
           "nn/__init__.py",
           "nn/loss.py",
           "nn/utils.py",
           "nn/core.py",
           "nn/io.py",
           "make_docs.py",
           "__init__.py"]

doc_url = "https://flaport.github.io/sax/"

git_url = "https://github.com/flaport/sax/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/08_backends.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['circuit_backends']

# Internal Cell
#nbdev_comment from __future__ import annotations

try:
    import jax
    JAX_AVAILABLE = True
except ImportError:
    JAX_AVAILABLE = False

try:
    import klujax
    KLUJAX_AVAILABLE = True
except ImportError:
    KLUJAX_AVAILABLE = False

from .default import evaluate_circuit
from .klu import evaluate_circuit_klu
from .additive import evaluate_circuit_additive

# Cell

circuit_backends = {
    "default": evaluate_circuit,
    "klu": evaluate_circuit_klu,
    "additive": evaluate_circuit_additive,
}

if (not JAX_AVAILABLE) or (not KLUJAX_AVAILABLE):
    del circuit_backends["klu"]


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06b_netlist_cleaning.ipynb (unless otherwise specified).

__all__ = ['remove_unused_instances']

# Cell
from collections import defaultdict
from typing import Dict, List

from natsort import natsorted
from .backends.additive import graph_edges
from .typing_ import SDict
import networkx as nx

try:
    import jax.numpy as jnp
    JAX_AVAILABLE = True
except ImportError:
    import numpy as jnp
    JAX_AVAILABLE = False

# Cell
def _get_connectivity_netlist(netlist):
    connectivity_netlist = {
        "instances": natsorted(netlist["instances"]),
        "connections": [
            (c1.split(",")[0], c2.split(",")[0])
            for c1, c2 in netlist["connections"].items()
        ],
        "ports": [(p, c.split(",")[0]) for p, c in netlist["ports"].items()],
    }
    return connectivity_netlist

# Cell
def _get_connectivity_graph(netlist):
    graph = nx.Graph()
    connectivity_netlist = _get_connectivity_netlist(netlist)
    for name in connectivity_netlist["instances"]:
        graph.add_node(name)
    for c1, c2 in connectivity_netlist["connections"]:
        graph.add_edge(c1, c2)
    for c1, c2 in connectivity_netlist["ports"]:
        graph.add_edge(c1, c2)
    return graph

# Cell
def _get_nodes_to_remove(graph, netlist):
    nodes = set()
    for port in netlist["ports"]:
        nodes |= nx.descendants(graph, port)
    nodes_to_remove = set(graph.nodes) - nodes
    return list(nodes_to_remove)

# Cell
def _remove_unused_instances_flat(flat_netlist):
    flat_netlist = {**flat_netlist}

    flat_netlist["instances"] = {**flat_netlist["instances"]}
    flat_netlist["connections"] = {**flat_netlist["connections"]}
    flat_netlist["ports"] = {**flat_netlist["ports"]}

    graph = _get_connectivity_graph(flat_netlist)
    names = set(_get_nodes_to_remove(graph, flat_netlist))

    for name in list(names):
        if name in flat_netlist["instances"]:
            del flat_netlist["instances"][name]

    for conn1, conn2 in list(flat_netlist["connections"].items()):
        for conn in [conn1, conn2]:
            name, _ = conn.split(",")
            if name in names and conn1 in flat_netlist["connections"]:
                del flat_netlist["connections"][conn1]

    for pname, conn in list(flat_netlist["ports"].items()):
        name, _ = conn.split(",")
        if name in names and pname in flat_netlist["ports"]:
            del flat_netlist["ports"][pname]

    return flat_netlist

# Cell
def remove_unused_instances(recursive_netlist):
    recursive_netlist = {**recursive_netlist}

    for name, flat_netlist in recursive_netlist.items():
        recursive_netlist[name] = _remove_unused_instances_flat(flat_netlist)

    return recursive_netlist